from operator import attrgetter
from typing import Callable

from graphql.language import (
    ast,
    Node,
    parse_value,
    is_definition_node,
    is_executable_definition_node,
    is_selection_node,
    is_value_node,
    is_const_value_node,
    is_type_node,
    is_type_system_definition_node,
    is_type_definition_node,
    is_type_system_extension_node,
    is_type_extension_node,
)

all_ast_nodes = sorted(
    [
        node_type()
        for node_type in vars(ast).values()
        if type(node_type) is type
        and issubclass(node_type, Node)
        and not node_type.__name__.startswith("Const")
    ],
    key=attrgetter("kind"),
)


def filter_nodes(predicate: Callable[[Node], bool]):
    return [node.kind for node in all_ast_nodes if predicate(node)]


def describe_ast_node_predicates():
    def check_definition_node():
        assert filter_nodes(is_definition_node) == [
            "definition",
            "directive_definition",
            "enum_type_definition",
            "enum_type_extension",
            "enum_value_definition",
            "executable_definition",
            "field_definition",
            "fragment_definition",
            "input_object_type_definition",
            "input_object_type_extension",
            "input_value_definition",
            "interface_type_definition",
            "interface_type_extension",
            "object_type_definition",
            "object_type_extension",
            "operation_definition",
            "scalar_type_definition",
            "scalar_type_extension",
            "schema_definition",
            "type_definition",
            "type_extension",
            "type_system_definition",
            "union_type_definition",
            "union_type_extension",
        ]

    def check_executable_definition_node():
        assert filter_nodes(is_executable_definition_node) == [
            "executable_definition",
            "fragment_definition",
            "operation_definition",
        ]

    def check_selection_node():
        assert filter_nodes(is_selection_node) == [
            "field",
            "fragment_spread",
            "inline_fragment",
            "selection",
        ]

    def check_value_node():
        assert filter_nodes(is_value_node) == [
            "boolean_value",
            "enum_value",
            "float_value",
            "int_value",
            "list_value",
            "null_value",
            "object_value",
            "string_value",
            "value",
            "variable",
        ]

    def check_const_value_node():
        assert is_const_value_node(parse_value('"value"')) is True
        assert is_const_value_node(parse_value("$var")) is False

        assert is_const_value_node(parse_value('{ field: "value" }')) is True
        assert is_const_value_node(parse_value("{ field: $var }")) is False

        assert is_const_value_node(parse_value('[ "value" ]')) is True
        assert is_const_value_node(parse_value("[ $var ]")) is False

    def check_type_node():
        assert filter_nodes(is_type_node) == [
            "list_type",
            "named_type",
            "non_null_type",
            "type",
        ]

    def check_type_system_definition_node():
        assert filter_nodes(is_type_system_definition_node) == [
            "directive_definition",
            "enum_type_definition",
            "enum_type_extension",
            "input_object_type_definition",
            "input_object_type_extension",
            "interface_type_definition",
            "interface_type_extension",
            "object_type_definition",
            "object_type_extension",
            "scalar_type_definition",
            "scalar_type_extension",
            "schema_definition",
            "type_definition",
            "type_extension",
            "type_system_definition",
            "union_type_definition",
            "union_type_extension",
        ]

    def check_type_definition_node():
        assert filter_nodes(is_type_definition_node) == [
            "enum_type_definition",
            "input_object_type_definition",
            "interface_type_definition",
            "object_type_definition",
            "scalar_type_definition",
            "type_definition",
            "union_type_definition",
        ]

    def check_type_system_extension_node():
        assert filter_nodes(is_type_system_extension_node) == [
            "enum_type_extension",
            "input_object_type_extension",
            "interface_type_extension",
            "object_type_extension",
            "scalar_type_extension",
            "schema_extension",
            "type_extension",
            "union_type_extension",
        ]

    def check_type_extension_node():
        assert filter_nodes(is_type_extension_node) == [
            "enum_type_extension",
            "input_object_type_extension",
            "interface_type_extension",
            "object_type_extension",
            "scalar_type_extension",
            "type_extension",
            "union_type_extension",
        ]



from pytest import mark, raises

from graphql.error import GraphQLError
from graphql.type import assert_name, assert_enum_value_name


def describe_assert_name():
    def pass_through_valid_name():
        assert assert_name("_ValidName123") == "_ValidName123"

    def throws_for_non_strings():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            assert_name({})  # type: ignore
        msg = str(exc_info.value)
        assert msg == "Expected name to be a string."

    def throws_on_empty_strings():
        with raises(GraphQLError) as exc_info:
            assert_name("")
        msg = str(exc_info.value)
        assert msg == "Expected name to be a non-empty string."

    def throws_for_names_with_invalid_characters():
        with raises(GraphQLError) as exc_info:
            assert_name(">--()-->")
        msg = str(exc_info.value)
        assert msg == "Names must only contain [_a-zA-Z0-9] but '>--()-->' does not."

    def throws_for_names_starting_with_invalid_characters():
        with raises(GraphQLError) as exc_info:
            assert_name("42MeaningsOfLife")
        msg = str(exc_info.value)
        assert msg == (
            "Names must start with [_a-zA-Z] but '42MeaningsOfLife' does not."
        )


def describe_assert_enum_value_name():
    def pass_through_valid_name():
        assert assert_enum_value_name("_ValidName123") == "_ValidName123"

    def throws_for_non_strings():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            assert_enum_value_name({})  # type: ignore
        msg = str(exc_info.value)
        assert msg == "Expected name to be a string."

    def throws_on_empty_strings():
        with raises(GraphQLError) as exc_info:
            assert_enum_value_name("")
        msg = str(exc_info.value)
        assert msg == "Expected name to be a non-empty string."

    def throws_for_names_with_invalid_characters():
        with raises(GraphQLError) as exc_info:
            assert_enum_value_name(">--()-->")
        msg = str(exc_info.value)
        assert msg == "Names must only contain [_a-zA-Z0-9] but '>--()-->' does not."

    def throws_for_names_starting_with_invalid_characters():
        with raises(GraphQLError) as exc_info:
            assert_enum_value_name("42MeaningsOfLife")
        msg = str(exc_info.value)
        assert msg == (
            "Names must start with [_a-zA-Z] but '42MeaningsOfLife' does not."
        )

    @mark.parametrize("name", ("true", "false", "null"))
    def throws_for_restricted_names(name):
        with raises(GraphQLError) as exc_info:
            assert_enum_value_name(name)
        msg = str(exc_info.value)
        assert msg == (f"Enum values cannot be named: {name}.")



"""Benchmarks for graphql

Benchmarks are disabled (only executed as tests) by default in setup.cfg.
You can enable them with --benchmark-enable if your want to execute them.

E.g. in order to execute all the benchmarks with tox using Python 3.9::

    tox -e py39 -- -k benchmarks --benchmark-enable
"""



from typing import Any, Dict

from ...error import GraphQLError
from ...language import DirectiveDefinitionNode, NameNode, VisitorAction, SKIP
from . import SDLValidationContext, SDLValidationRule

__all__ = ["UniqueDirectiveNamesRule"]


class UniqueDirectiveNamesRule(SDLValidationRule):
    """Unique directive names

    A GraphQL document is only valid if all defined directives have unique names.
    """

    def __init__(self, context: SDLValidationContext):
        super().__init__(context)
        self.known_directive_names: Dict[str, NameNode] = {}
        self.schema = context.schema

    def enter_directive_definition(
        self, node: DirectiveDefinitionNode, *_args: Any
    ) -> VisitorAction:
        directive_name = node.name.value

        if self.schema and self.schema.get_directive(directive_name):
            self.report_error(
                GraphQLError(
                    f"Directive '@{directive_name}' already exists in the schema."
                    " It cannot be redefined.",
                    node.name,
                )
            )
        else:
            if directive_name in self.known_directive_names:
                self.report_error(
                    GraphQLError(
                        f"There can be only one directive named '@{directive_name}'.",
                        [self.known_directive_names[directive_name], node.name],
                    )
                )
            else:
                self.known_directive_names[directive_name] = node.name
            return SKIP

        return None



from typing import TYPE_CHECKING

from .graphql_error import GraphQLError

if TYPE_CHECKING:
    from ..language.source import Source  # noqa: F401

__all__ = ["GraphQLSyntaxError"]


class GraphQLSyntaxError(GraphQLError):
    """A GraphQLError representing a syntax error."""

    def __init__(self, source: "Source", position: int, description: str) -> None:
        super().__init__(
            f"Syntax Error: {description}", source=source, positions=[position]
        )
        self.description = description



from copy import deepcopy
from typing import Dict, List, TypeVar

from .frozen_error import FrozenError

__all__ = ["FrozenList"]


T = TypeVar("T", covariant=True)


class FrozenList(List[T]):
    """List that can only be read, but not changed.

    .. deprecated:: 3.2
       Use tuples or lists and the Collection type instead. Will be removed in v3.3.
    """

    def __delitem__(self, key):
        raise FrozenError

    def __setitem__(self, key, value):
        raise FrozenError

    def __add__(self, value):
        if isinstance(value, tuple):
            value = list(value)
        return list.__add__(self, value)

    def __iadd__(self, value):
        raise FrozenError

    def __mul__(self, value):
        return list.__mul__(self, value)

    def __imul__(self, value):
        raise FrozenError

    def __hash__(self):
        return hash(tuple(self))

    def __copy__(self) -> "FrozenList":
        return FrozenList(self)

    def __deepcopy__(self, memo: Dict) -> "FrozenList":
        return FrozenList(deepcopy(value, memo) for value in self)

    def append(self, x):
        raise FrozenError

    def extend(self, iterable):
        raise FrozenError

    def insert(self, i, x):
        raise FrozenError

    def remove(self, x):
        raise FrozenError

    def pop(self, i=None):
        raise FrozenError

    def clear(self):
        raise FrozenError

    def sort(self, *, key=None, reverse=False):
        raise FrozenError

    def reverse(self):
        raise FrozenError



from functools import partial

from graphql.pyutils import Undefined, inspect
from graphql.type import (
    GraphQLArgument,
    GraphQLField,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLScalarType,
    GraphQLString,
)
from graphql.validation import ValuesOfCorrectTypeRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, ValuesOfCorrectTypeRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_values_of_correct_type():
    def describe_valid_values():
        def good_int_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: 2)
                  }
                }
                """
            )

        def good_negative_int_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: -2)
                  }
                }
                """
            )

        def good_boolean_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    booleanArgField(intArg: true)
                  }
                }
                """
            )

        def good_string_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    stringArgField(intArg: "foo")
                  }
                }
                """
            )

        def good_float_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    floatArgField(intArg: 1.1)
                  }
                }
                """
            )

        def good_negative_float_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    floatArgField(intArg: -1.1)
                  }
                }
                """
            )

        def int_into_id():
            assert_valid(
                """
                {
                  complicatedArgs {
                    idArgField(idArg: 1)
                  }
                }
                """
            )

        def string_into_id():
            assert_valid(
                """
                {
                  complicatedArgs {
                    idArgField(idArg: "someIdString")
                  }
                }
                """
            )

        def good_enum_value():
            assert_valid(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: SIT)
                  }
                }
                """
            )

        def enum_with_undefined_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    enumArgField(enumArg: UNKNOWN)
                  }
                }
                """
            )

        def enum_with_null_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    enumArgField(enumArg: NO_FUR)
                  }
                }
                """
            )

        def null_into_nullable_type():
            assert_valid(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: null)
                  }
                }
                """
            )

            assert_valid(
                """
                {
                  dog(a: null, b: null, c:{ requiredField: true, intField: null }) {
                    name
                  }
                }
                """
            )

    def describe_invalid_string_values():
        def int_into_string():
            assert_errors(
                """
                {
                  complicatedArgs {
                    stringArgField(stringArg: 1)
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: 1",
                        "locations": [(4, 47)],
                    },
                ],
            )

        def float_into_string():
            assert_errors(
                """
                {
                  complicatedArgs {
                    stringArgField(stringArg: 1.0)
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: 1.0",
                        "locations": [(4, 47)],
                    },
                ],
            )

        def boolean_into_string():
            assert_errors(
                """
                {
                  complicatedArgs {
                    stringArgField(stringArg: true)
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: true",
                        "locations": [(4, 47)],
                    },
                ],
            )

        def unquoted_string_into_string():
            assert_errors(
                """
                {
                  complicatedArgs {
                    stringArgField(stringArg: BAR)
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: BAR",
                        "locations": [(4, 47)],
                    },
                ],
            )

    def describe_invalid_int_values():
        def string_into_int():
            assert_errors(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: "3")
                  }
                }
                """,
                [
                    {
                        "message": 'Int cannot represent non-integer value: "3"',
                        "locations": [(4, 41)],
                    },
                ],
            )

        def big_int_into_int():
            assert_errors(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: 829384293849283498239482938)
                  }
                }
                """,
                [
                    {
                        "message": "Int cannot represent non 32-bit signed integer"
                        " value: 829384293849283498239482938",
                        "locations": [(4, 41)],
                    },
                ],
            )

        def unquoted_string_into_int():
            assert_errors(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: FOO)
                  }
                }
                """,
                [
                    {
                        "message": "Int cannot represent non-integer value: FOO",
                        "locations": [(4, 41)],
                    },
                ],
            )

        def simple_float_into_int():
            assert_errors(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: 3.0)
                  }
                }
                """,
                [
                    {
                        "message": "Int cannot represent non-integer value: 3.0",
                        "locations": [(4, 41)],
                    }
                ],
            )

        def float_into_int():
            assert_errors(
                """
                {
                  complicatedArgs {
                    intArgField(intArg: 3.333)
                  }
                }
                """,
                [
                    {
                        "message": "Int cannot represent non-integer value: 3.333",
                        "locations": [(4, 41)],
                    },
                ],
            )

    def describe_invalid_float_values():
        def string_into_float():
            assert_errors(
                """
                {
                  complicatedArgs {
                    floatArgField(floatArg: "3.333")
                  }
                }
                """,
                [
                    {
                        "message": 'Float cannot represent non numeric value: "3.333"',
                        "locations": [(4, 45)],
                    },
                ],
            )

        def boolean_into_float():
            assert_errors(
                """
                {
                  complicatedArgs {
                    floatArgField(floatArg: true)
                  }
                }
                """,
                [
                    {
                        "message": "Float cannot represent non numeric value: true",
                        "locations": [(4, 45)],
                    },
                ],
            )

        def unquoted_into_float():
            assert_errors(
                """
                {
                  complicatedArgs {
                    floatArgField(floatArg: FOO)
                  }
                }
                """,
                [
                    {
                        "message": "Float cannot represent non numeric value: FOO",
                        "locations": [(4, 45)],
                    },
                ],
            )

    def describe_invalid_boolean_value():
        def int_into_boolean():
            assert_errors(
                """
                {
                  complicatedArgs {
                    booleanArgField(booleanArg: 2)
                  }
                }
                """,
                [
                    {
                        "message": "Boolean cannot represent a non boolean value: 2",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def float_into_boolean():
            assert_errors(
                """
                {
                  complicatedArgs {
                    booleanArgField(booleanArg: 1.0)
                  }
                }
                """,
                [
                    {
                        "message": "Boolean cannot represent a non boolean value: 1.0",
                        "locations": [(4, 49)],
                    }
                ],
            )

        def string_into_boolean():
            assert_errors(
                """
                {
                  complicatedArgs {
                    booleanArgField(booleanArg: "true")
                  }
                }
                """,
                [
                    {
                        "message": "Boolean cannot represent a non boolean value:"
                        ' "true"',
                        "locations": [(4, 49)],
                    }
                ],
            )

        def unquoted_into_boolean():
            assert_errors(
                """
                {
                  complicatedArgs {
                    booleanArgField(booleanArg: TRUE)
                  }
                }
                """,
                [
                    {
                        "message": "Boolean cannot represent a non boolean value: TRUE",
                        "locations": [(4, 49)],
                    },
                ],
            )

    def describe_invalid_id_value():
        def float_into_id():
            assert_errors(
                """
                {
                  complicatedArgs {
                    idArgField(idArg: 1.0)
                  }
                }
                """,
                [
                    {
                        "message": "ID cannot represent a non-string"
                        " and non-integer value: 1.0",
                        "locations": [(4, 39)],
                    }
                ],
            )

        def boolean_into_id():
            assert_errors(
                """
                {
                  complicatedArgs {
                    idArgField(idArg: true)
                  }
                }
                """,
                [
                    {
                        "message": "ID cannot represent a non-string"
                        " and non-integer value: true",
                        "locations": [(4, 39)],
                    },
                ],
            )

        def unquoted_into_id():
            assert_errors(
                """
                {
                  complicatedArgs {
                    idArgField(idArg: SOMETHING)
                  }
                }
                """,
                [
                    {
                        "message": "ID cannot represent a non-string"
                        " and non-integer value: SOMETHING",
                        "locations": [(4, 39)],
                    },
                ],
            )

    def describe_invalid_enum_value():
        def int_into_enum():
            assert_errors(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: 2)
                  }
                }
                """,
                [
                    {
                        "message": "Enum 'DogCommand' cannot represent non-enum value:"
                        " 2.",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def float_into_enum():
            assert_errors(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: 1.0)
                  }
                }
                """,
                [
                    {
                        "message": "Enum 'DogCommand' cannot represent non-enum value:"
                        " 1.0.",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def string_into_enum():
            assert_errors(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: "SIT")
                  }
                }
                """,
                [
                    {
                        "message": "Enum 'DogCommand' cannot represent non-enum value:"
                        ' "SIT".'
                        " Did you mean the enum value 'SIT'?",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def boolean_into_enum():
            assert_errors(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: true)
                  }
                }
                """,
                [
                    {
                        "message": "Enum 'DogCommand' cannot represent non-enum value:"
                        " true.",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def unknown_enum_value_into_enum():
            assert_errors(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: JUGGLE)
                  }
                }
                """,
                [
                    {
                        "message": "Value 'JUGGLE'"
                        " does not exist in 'DogCommand' enum.",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def different_case_enum_value_into_enum():
            assert_errors(
                """
                {
                  dog {
                    doesKnowCommand(dogCommand: sit)
                  }
                }
                """,
                [
                    {
                        "message": "Value 'sit' does not exist in 'DogCommand' enum."
                        " Did you mean the enum value 'SIT'?",
                        "locations": [(4, 49)],
                    },
                ],
            )

    def describe_valid_list_value():
        def good_list_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    stringListArgField(stringListArg: ["one", null, "two"])
                  }
                }
                """
            )

        def empty_list_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    stringListArgField(stringListArg: [])
                  }
                }
                """
            )

        def null_value():
            assert_valid(
                """
                {
                  complicatedArgs {
                    stringListArgField(stringListArg: null)
                  }
                }
                """
            )

        def single_value_into_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    stringListArgField(stringListArg: "one")
                  }
                }
                """
            )

    def describe_invalid_list_value():
        def incorrect_item_type():
            assert_errors(
                """
                {
                  complicatedArgs {
                    stringListArgField(stringListArg: ["one", 2])
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: 2",
                        "locations": [(4, 63)],
                    },
                ],
            )

        def single_value_of_incorrect_type():
            assert_errors(
                """
                {
                  complicatedArgs {
                    stringListArgField(stringListArg: 1)
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: 1",
                        "locations": [(4, 55)],
                    },
                ],
            )

    def describe_valid_non_nullable_value():
        def arg_on_optional_arg():
            assert_valid(
                """
                {
                  dog {
                    isHouseTrained(atOtherHomes: true)
                  }
                }
                """
            )

        def no_arg_on_optional_arg():
            assert_valid(
                """
                {
                  dog {
                    isHouseTrained
                  }
                }
                """
            )

        def multiple_args():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleReqs(req1: 1, req2: 2)
                  }
                }
                """
            )

        def multiple_args_reverse_order():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleReqs(req2: 2, req1: 1)
                  }
                }
                """
            )

        def no_args_on_multiple_optional():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOpts
                  }
                }
                """
            )

        def one_arg_on_multiple_optional():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOpts(opt1: 1)
                  }
                }
                """
            )

        def second_arg_on_multiple_optional():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOpts(opt2: 1)
                  }
                }
                """
            )

        def multiple_required_args_on_mixed_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOptAndReq(req1: 3, req2: 4)
                  }
                }
                """
            )

        def multiple_required_and_one_optional_arg_on_mixed_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOptAndReq(req1: 3, req2: 4, opt1: 5)
                  }
                }
                """
            )

        def all_required_and_optional_args_on_mixed_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOptAndReq(req1: 3, req2: 4, opt1: 5, opt2: 6)
                  }
                }
                """
            )

    def describe_invalid_non_nullable_value():
        def incorrect_value_type():
            assert_errors(
                """
                {
                  complicatedArgs {
                    multipleReqs(req2: "two", req1: "one")
                  }
                }
                """,
                [
                    {
                        "message": 'Int cannot represent non-integer value: "two"',
                        "locations": [(4, 40)],
                    },
                    {
                        "message": 'Int cannot represent non-integer value: "one"',
                        "locations": [(4, 53)],
                    },
                ],
            )

        def incorrect_value_and_missing_argument_provided_required_arguments():
            assert_errors(
                """
                {
                  complicatedArgs {
                    multipleReqs(req1: "one")
                  }
                }
                """,
                [
                    {
                        "message": 'Int cannot represent non-integer value: "one"',
                        "locations": [(4, 40)],
                    },
                ],
            )

        def null_value():
            assert_errors(
                """
                {
                  complicatedArgs {
                    multipleReqs(req1: null)
                  }
                }
                """,
                [
                    {
                        "message": "Expected value of type 'Int!', found null.",
                        "locations": [(4, 40)],
                    },
                ],
            )

    def describe_valid_input_object_value():
        def optional_arg_despite_required_field_in_type():
            assert_valid(
                """
                {
                  complicatedArgs {
                    complexArgField
                  }
                }
                """
            )

        def partial_object_only_required():
            assert_valid(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: { requiredField: true })
                  }
                }
                """
            )

        def partial_object_required_field_can_be_falsy():
            assert_valid(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: { requiredField: false })
                  }
                }
                """
            )

        def partial_object_including_required():
            assert_valid(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: { requiredField: true, intField: 4 })
                  }
                }
                """
            )

        def full_object():
            assert_valid(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: {
                      requiredField: true,
                      intField: 4,
                      stringField: "foo",
                      booleanField: false,
                      stringListField: ["one", "two"]
                    })
                  }
                }
                """
            )

        def full_object_with_fields_in_different_order():
            assert_valid(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: {
                      stringListField: ["one", "two"],
                      booleanField: false,
                      requiredField: true,
                      stringField: "foo",
                      intField: 4,
                    })
                  }
                }
                """
            )

    def describe_invalid_input_object_value():
        def partial_object_missing_required():
            assert_errors(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: { intField: 4 })
                  }
                }
                """,
                [
                    {
                        "message": "Field 'ComplexInput.requiredField'"
                        " of required type 'Boolean!' was not provided.",
                        "locations": [(4, 49)],
                    },
                ],
            )

        def partial_object_invalid_field_type():
            assert_errors(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: {
                      stringListField: ["one", 2],
                      requiredField: true,
                    })
                  }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: 2",
                        "locations": [(5, 48)],
                    },
                ],
            )

        def partial_object_null_to_non_null_field():
            assert_errors(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: {
                      requiredField: true,
                      nonNullField: null,
                    })
                  }
                }
                """,
                [
                    {
                        "message": "Expected value of type 'Boolean!', found null.",
                        "locations": [(6, 37)],
                    }
                ],
            )

        def partial_object_unknown_field_arg():
            assert_errors(
                """
                {
                  complicatedArgs {
                    complexArgField(complexArg: {
                      requiredField: true,
                      invalidField: "value"
                    })
                  }
                }
                """,
                [
                    {
                        "message": "Field 'invalidField'"
                        " is not defined by type 'ComplexInput'."
                        " Did you mean 'intField'?",
                        "locations": [(6, 23)],
                    },
                ],
            )

        def reports_original_error_for_custom_scalar_which_throws():
            def parse_value(value):
                raise Exception(f"Invalid scalar is always invalid: {inspect(value)}")

            custom_scalar = GraphQLScalarType("Invalid", parse_value=parse_value)

            schema = GraphQLSchema(
                query=GraphQLObjectType(
                    "Query",
                    {
                        "invalidArg": GraphQLField(
                            GraphQLString, {"arg": GraphQLArgument(custom_scalar)}
                        )
                    },
                )
            )

            errors = assert_errors(
                "{ invalidArg(arg: 123) }",
                [
                    {
                        "message": "Expected value of type 'Invalid', found 123;"
                        " Invalid scalar is always invalid: 123",
                        "locations": [(1, 19)],
                    }
                ],
                schema=schema,
            )

            assert str(errors[0].original_error) == (
                "Invalid scalar is always invalid: 123"
            )

        def reports_error_for_custom_scalar_that_returns_undefined():
            custom_scalar = GraphQLScalarType(
                "CustomScalar", parse_value=lambda value: Undefined
            )

            schema = GraphQLSchema(
                GraphQLObjectType(
                    "Query",
                    {
                        "invalidArg": GraphQLField(
                            GraphQLString, args={"arg": GraphQLArgument(custom_scalar)}
                        )
                    },
                )
            )

            assert_errors(
                "{ invalidArg(arg: 123) }",
                [
                    {
                        "message": "Expected value of type 'CustomScalar', found 123.",
                        "locations": [(1, 19)],
                    },
                ],
                schema=schema,
            )

        def allows_custom_scalar_to_accept_complex_literals():
            custom_scalar = GraphQLScalarType("Any")
            schema = GraphQLSchema(
                query=GraphQLObjectType(
                    "Query",
                    {
                        "anyArg": GraphQLField(
                            GraphQLString, {"arg": GraphQLArgument(custom_scalar)}
                        )
                    },
                )
            )

            assert_valid(
                """
                {
                  test1: anyArg(arg: 123)
                  test2: anyArg(arg: "abc")
                  test3: anyArg(arg: [123, "abc"])
                  test4: anyArg(arg: {deep: [123, "abc"]})
                }
                """,
                schema=schema,
            )

    def describe_directive_arguments():
        def with_directives_of_valid_types():
            assert_valid(
                """
                {
                  dog @include(if: true) {
                    name
                  }
                  human @skip(if: false) {
                    name
                  }
                }
                """
            )

        def with_directives_with_incorrect_types():
            assert_errors(
                """
                {
                  dog @include(if: "yes") {
                    name @skip(if: ENUM)
                  }
                }
                """,
                [
                    {
                        "message": "Boolean cannot represent a non boolean value:"
                        ' "yes"',
                        "locations": [(3, 36)],
                    },
                    {
                        "message": "Boolean cannot represent a non boolean value: ENUM",
                        "locations": [(4, 36)],
                    },
                ],
            )

    def describe_variable_default_values():
        def variables_with_valid_default_values():
            assert_valid(
                """
                query WithDefaultValues(
                  $a: Int = 1,
                  $b: String = "ok",
                  $c: ComplexInput = { requiredField: true, intField: 3 }
                  $d: Int! = 123
                ) {
                  dog { name }
                }
                """
            )

        def variables_with_valid_default_null_values():
            assert_valid(
                """
                query WithDefaultValues(
                  $a: Int = null,
                  $b: String = null,
                  $c: ComplexInput = { requiredField: true, intField: null }
                ) {
                  dog { name }
                }
                """
            )

        def variables_with_invalid_default_null_values():
            assert_errors(
                """
                query WithDefaultValues(
                  $a: Int! = null,
                  $b: String! = null,
                  $c: ComplexInput = { requiredField: null, intField: null }
                ) {
                  dog { name }
                }
                """,
                [
                    {
                        "message": "Expected value of type 'Int!', found null.",
                        "locations": [(3, 30)],
                    },
                    {
                        "message": "Expected value of type 'String!', found null.",
                        "locations": [(4, 33)],
                    },
                    {
                        "message": "Expected value of type 'Boolean!', found null.",
                        "locations": [(5, 55)],
                    },
                ],
            )

        def variables_with_invalid_default_values():
            assert_errors(
                """
                query InvalidDefaultValues(
                  $a: Int = "one",
                  $b: String = 4,
                  $c: ComplexInput = "NotVeryComplex"
                ) {
                  dog { name }
                }
                """,
                [
                    {
                        "message": 'Int cannot represent non-integer value: "one"',
                        "locations": [(3, 29)],
                    },
                    {
                        "message": "String cannot represent a non string value: 4",
                        "locations": [(4, 32)],
                    },
                    {
                        "message": "Expected value of type 'ComplexInput',"
                        ' found "NotVeryComplex".',
                        "locations": [(5, 38)],
                    },
                ],
            )

        def variables_with_complex_invalid_default_values():
            assert_errors(
                """
                query WithDefaultValues(
                  $a: ComplexInput = { requiredField: 123, intField: "abc" }
                ) {
                  dog { name }
                }
                """,
                [
                    {
                        "message": "Boolean cannot represent a non boolean value: 123",
                        "locations": [(3, 55)],
                    },
                    {
                        "message": 'Int cannot represent non-integer value: "abc"',
                        "locations": [(3, 70)],
                    },
                ],
            )

        def complex_variables_missing_required_fields():
            assert_errors(
                """
                query MissingRequiredField($a: ComplexInput = {intField: 3}) {
                  dog { name }
                }
                """,
                [
                    {
                        "message": "Field 'ComplexInput.requiredField'"
                        " of required type 'Boolean!' was not provided.",
                        "locations": [(2, 63)],
                    },
                ],
            )

        def list_variables_with_invalid_item():
            assert_errors(
                """
                query InvalidItem($a: [String] = ["one", 2]) {
                  dog { name }
                }
                """,
                [
                    {
                        "message": "String cannot represent a non string value: 2",
                        "locations": [(2, 58)],
                    },
                ],
            )



from graphql import build_schema, parse, validate

from ..fixtures import big_schema_sdl  # noqa: F401


def test_validate_invalid_query(benchmark, big_schema_sdl):  # noqa: F811
    schema = build_schema(big_schema_sdl, assume_valid=True)
    query_ast = parse(
        """
        {
          unknownField
          ... on unknownType {
            anotherUnknownField
            ...unknownFragment
          }
        }

        fragment TestFragment on anotherUnknownType {
          yetAnotherUnknownField
        }
        """
    )
    result = benchmark(lambda: validate(schema, query_ast))
    assert result == [
        {
            "message": "Cannot query field 'unknownField' on type 'Query'.",
            "locations": [(3, 11)],
        },
        {"message": "Unknown type 'unknownType'.", "locations": [(4, 18)]},
        {"message": "Unknown fragment 'unknownFragment'.", "locations": [(6, 16)]},
        {"message": "Unknown type 'anotherUnknownType'.", "locations": [(10, 34)]},
        {"message": "Fragment 'TestFragment' is never used.", "locations": [(10, 9)]},
    ]



from typing import TYPE_CHECKING, Collection, Optional, Union

from ..pyutils import inspect
from .graphql_error import GraphQLError

if TYPE_CHECKING:
    from ..language.ast import Node  # noqa: F401

__all__ = ["located_error"]


def located_error(
    original_error: Exception,
    nodes: Optional[Union["None", Collection["Node"]]] = None,
    path: Optional[Collection[Union[str, int]]] = None,
) -> GraphQLError:
    """Located GraphQL Error

    Given an arbitrary Exception, presumably thrown while attempting to execute a
    GraphQL operation, produce a new GraphQLError aware of the location in the document
    responsible for the original Exception.
    """
    # Sometimes a non-error is thrown, wrap it as a TypeError to ensure consistency.
    if not isinstance(original_error, Exception):
        original_error = TypeError(f"Unexpected error value: {inspect(original_error)}")
    # Note: this uses a brand-check to support GraphQL errors originating from
    # other contexts.
    if isinstance(original_error, GraphQLError) and original_error.path is not None:
        return original_error
    try:
        # noinspection PyUnresolvedReferences
        message = str(original_error.message)  # type: ignore
    except AttributeError:
        message = str(original_error)
    try:
        # noinspection PyUnresolvedReferences
        source = original_error.source  # type: ignore
    except AttributeError:
        source = None
    try:
        # noinspection PyUnresolvedReferences
        positions = original_error.positions  # type: ignore
    except AttributeError:
        positions = None
    try:
        # noinspection PyUnresolvedReferences
        nodes = original_error.nodes or nodes  # type: ignore
    except AttributeError:
        pass
    return GraphQLError(message, nodes, source, positions, path, original_error)



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import OverlappingFieldsCanBeMergedRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, OverlappingFieldsCanBeMergedRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_overlapping_fields_can_be_merged():
    def unique_fields():
        assert_valid(
            """
            fragment uniqueFields on Dog {
              name
              nickname
            }
            """
        )

    def identical_fields():
        assert_valid(
            """
            fragment mergeIdenticalFields on Dog {
              name
              name
            }
            """
        )

    def identical_fields_with_identical_args():
        assert_valid(
            """
            fragment mergeIdenticalFieldsWithIdenticalArgs on Dog {
              doesKnowCommand(dogCommand: SIT)
              doesKnowCommand(dogCommand: SIT)
            }
            """
        )

    def identical_fields_with_identical_directives():
        assert_valid(
            """
            fragment mergeSameFieldsWithSameDirectives on Dog {
              name @include(if: true)
              name @include(if: true)
            }
            """
        )

    def different_args_with_different_aliases():
        assert_valid(
            """
            fragment differentArgsWithDifferentAliases on Dog {
              knowsSit: doesKnowCommand(dogCommand: SIT)
              knowsDown: doesKnowCommand(dogCommand: DOWN)
            }
            """
        )

    def different_directives_with_different_aliases():
        assert_valid(
            """
            fragment differentDirectivesWithDifferentAliases on Dog {
              nameIfTrue: name @include(if: true)
              nameIfFalse: name @include(if: false)
            }
            """
        )

    def different_skip_or_include_directives_accepted():
        # Note: Differing skip/include directives don't create an ambiguous
        # return value and are acceptable in conditions where differing runtime
        # values may have the same desired effect of including/skipping a field
        assert_valid(
            """
            fragment differentDirectivesWithDifferentAliases on Dog {
              name @include(if: true)
              name @include(if: false)
            }
            """
        )

    def same_aliases_with_different_field_targets():
        assert_errors(
            """
            fragment sameAliasesWithDifferentFieldTargets on Dog {
              fido: name
              fido: nickname
            }
            """,
            [
                {
                    "message": "Fields 'fido' conflict"
                    " because 'name' and 'nickname' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 15)],
                    "path": None,
                }
            ],
        )

    def same_aliases_allowed_on_non_overlapping_fields():
        assert_valid(
            """
            fragment sameAliasesWithDifferentFieldTargets on Pet {
              ... on Dog {
                name
              }
              ... on Cat {
                name: nickname
              }
            }
            """
        )

    def alias_masking_direct_field_access():
        assert_errors(
            """
            fragment aliasMaskingDirectFieldAccess on Dog {
              name: nickname
              name
            }
            """,
            [
                {
                    "message": "Fields 'name' conflict"
                    " because 'nickname' and 'name' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 15)],
                }
            ],
        )

    def different_args_second_adds_an_argument():
        assert_errors(
            """
            fragment conflictingArgs on Dog {
              doesKnowCommand
              doesKnowCommand(dogCommand: HEEL)
            }
            """,
            [
                {
                    "message": "Fields 'doesKnowCommand' conflict"
                    " because they have differing arguments."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 15)],
                }
            ],
        )

    def different_args_second_missing_an_argument():
        assert_errors(
            """
            fragment conflictingArgs on Dog {
              doesKnowCommand(dogCommand: SIT)
              doesKnowCommand
            }
            """,
            [
                {
                    "message": "Fields 'doesKnowCommand' conflict"
                    " because they have differing arguments."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 15)],
                }
            ],
        )

    def conflicting_arg_values():
        assert_errors(
            """
            fragment conflictingArgs on Dog {
              doesKnowCommand(dogCommand: SIT)
              doesKnowCommand(dogCommand: HEEL)
            }
            """,
            [
                {
                    "message": "Fields 'doesKnowCommand' conflict"
                    " because they have differing arguments."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 15)],
                }
            ],
        )

    def conflicting_arg_names():
        assert_errors(
            """
            fragment conflictingArgs on Dog {
              isAtLocation(x: 0)
              isAtLocation(y: 0)
            }
            """,
            [
                {
                    "message": "Fields 'isAtLocation' conflict"
                    " because they have differing arguments."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 15)],
                }
            ],
        )

    def allows_different_args_where_no_conflict_is_possible():
        # This is valid since no object can be both a "Dog" and a "Cat", thus
        # these fields can never overlap.
        assert_valid(
            """
            fragment conflictingArgs on Pet {
              ... on Dog {
                name(surname: true)
              }
              ... on Cat {
                name
              }
            }
            """
        )

    def allows_different_order_of_args():
        schema = build_schema(
            """
            type Query {
              someField(a: String, b: String): String
            }
            """
        )
        # This is valid since arguments are unordered, see:
        # https://spec.graphql.org/draft/#
        # sec-Language.Arguments.Arguments-are-unordered
        assert_valid(
            """
            {
              someField(a: null, b: null)
              someField(b: null, a: null)
            }
            """,
            schema=schema,
        )

    def allows_different_order_of_input_object_fields_in_arg_values():
        schema = build_schema(
            """
            input SomeInput {
              a: String
              b: String
            }

            type Query {
              someField(arg: SomeInput): String
            }
            """
        )
        # This is valid since input object fields are unordered, see:
        # https://spec.graphql.org/draft/#
        # sec-Input-Object-Values.Input-object-fields-are-unordered
        assert_valid(
            """
            {
              someField(arg: { a: null, b: null })
              someField(arg: { b: null, a: null })
            }
            """,
            schema=schema,
        )

    def encounters_conflict_in_fragments():
        assert_errors(
            """
            {
              ...A
              ...B
            }
            fragment A on Type {
              x: a
            }
            fragment B on Type {
              x: b
            }
            """,
            [
                {
                    "message": "Fields 'x' conflict"
                    " because 'a' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(7, 15), (10, 15)],
                }
            ],
        )

    def reports_each_conflict_once():
        assert_errors(
            """
            {
              f1 {
                ...A
                ...B
              }
              f2 {
                ...B
                ...A
              }
              f3 {
                ...A
                ...B
                x: c
              }
            }
            fragment A on Type {
              x: a
            }
            fragment B on Type {
              x: b
            }
            """,
            [
                {
                    "message": "Fields 'x' conflict"
                    " because 'a' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(18, 15), (21, 15)],
                },
                {
                    "message": "Fields 'x' conflict"
                    " because 'c' and 'a' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(14, 17), (18, 15)],
                },
                {
                    "message": "Fields 'x' conflict"
                    " because 'c' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(14, 17), (21, 15)],
                },
            ],
        )

    def deep_conflict():
        assert_errors(
            """
            {
              field {
                x: a
              },
              field {
                x: b
              }
            }
            """,
            [
                {
                    "message": "Fields 'field' conflict"
                    " because subfields 'x' conflict"
                    " because 'a' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 17), (6, 15), (7, 17)],
                }
            ],
        )

    def deep_conflict_with_multiple_issues():
        assert_errors(
            """
            {
              field {
                x: a
                y: c
              },
              field {
                x: b
                y: d
              }
            }
            """,
            [
                {
                    "message": "Fields 'field' conflict"
                    " because subfields 'x' conflict"
                    " because 'a' and 'b' are different fields"
                    " and subfields 'y' conflict"
                    " because 'c' and 'd' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (4, 17), (5, 17), (7, 15), (8, 17), (9, 17)],
                    "path": None,
                }
            ],
        )

    def very_deep_conflict():
        assert_errors(
            """
            {
              field {
                deepField {
                  x: a
                }
              },
              field {
                deepField {
                  x: b
                }
              }
            }
            """,
            [
                {
                    "message": "Fields 'field' conflict"
                    " because subfields 'deepField' conflict"
                    " because subfields 'x' conflict"
                    " because 'a' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [
                        (3, 15),
                        (4, 17),
                        (5, 19),
                        (8, 15),
                        (9, 17),
                        (10, 19),
                    ],
                    "path": None,
                }
            ],
        )

    def reports_deep_conflict_to_nearest_common_ancestor():
        assert_errors(
            """
            {
              field {
                deepField {
                  x: a
                }
                deepField {
                  x: b
                }
              },
              field {
                deepField {
                  y
                }
              }
            }
            """,
            [
                {
                    "message": "Fields 'deepField' conflict"
                    " because subfields 'x' conflict"
                    " because 'a' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(4, 17), (5, 19), (7, 17), (8, 19)],
                }
            ],
        )

    def reports_deep_conflict_to_nearest_common_ancestor_in_fragments():
        assert_errors(
            """
            {
              field {
                ...F
              }
              field {
                ...F
              }
            }
            fragment F on T {
              deepField {
                deeperField {
                  x: a
                }
                deeperField {
                  x: b
                }
              },
              deepField {
                deeperField {
                  y
                }
              }
            }
            """,
            [
                {
                    "message": "Fields 'deeperField' conflict"
                    " because subfields 'x' conflict"
                    " because 'a' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(12, 17), (13, 19), (15, 17), (16, 19)],
                }
            ],
        )

    def reports_deep_conflict_in_nested_fragments():
        assert_errors(
            """
            {
              field {
                ...F
              },
              field {
                ...I
              }
            }
            fragment F on T {
              x: a
              ...G
            }
            fragment G on T {
              y: c
            }
            fragment I on T {
              y: d
              ...J
            }
            fragment J on T {
              x: b
            }
            """,
            [
                {
                    "message": "Fields 'field' conflict"
                    " because subfields 'x' conflict"
                    " because 'a' and 'b' are different fields"
                    " and subfields 'y' conflict"
                    " because 'c' and 'd' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [
                        (3, 15),
                        (11, 15),
                        (15, 15),
                        (6, 15),
                        (22, 15),
                        (18, 15),
                    ],
                    "path": None,
                }
            ],
        )

    def ignores_unknown_fragments():
        assert_valid(
            """
            {
              field
              ...Unknown
              ...Known
            }

            fragment Known on T {
              field
              ...OtherUnknown
            }
            """
        )

    def describe_return_types_must_be_unambiguous():

        schema = build_schema(
            """
            interface SomeBox {
              deepBox: SomeBox
              unrelatedField: String
            }

            type StringBox implements SomeBox {
              scalar: String
              deepBox: StringBox
              unrelatedField: String
              listStringBox: [StringBox]
              stringBox: StringBox
              intBox: IntBox
            }

            type IntBox implements SomeBox {
              scalar: Int
              deepBox: IntBox
              unrelatedField: String
              listStringBox: [StringBox]
              stringBox: StringBox
              intBox: IntBox
            }

            interface NonNullStringBox1 {
              scalar: String!
            }

            type NonNullStringBox1Impl implements SomeBox & NonNullStringBox1 {
              scalar: String!
              unrelatedField: String
              deepBox: SomeBox
            }

            interface NonNullStringBox2 {
              scalar: String!
            }

            type NonNullStringBox2Impl implements SomeBox & NonNullStringBox2 {
              scalar: String!
              unrelatedField: String
              deepBox: SomeBox
            }

            type Connection {
              edges: [Edge]
            }

            type Edge {
              node: Node
            }

            type Node {
              id: ID
              name: String
            }

            type Query {
              someBox: SomeBox
              connection: Connection
            }
            """
        )

        def conflicting_return_types_which_potentially_overlap():
            # This is invalid since an object could potentially be both the
            # Object type IntBox and the interface type NonNullStringBox1.
            # While that condition does not exist in the current schema, the
            # schema could expand in the future to allow this.
            assert_errors(
                """
                {
                  someBox {
                    ...on IntBox {
                      scalar
                    }
                    ...on NonNullStringBox1 {
                      scalar
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'scalar' conflict because"
                        " they return conflicting types 'Int' and 'String!'."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(5, 23), (8, 23)],
                    }
                ],
                schema,
            )

        def compatible_return_shapes_on_different_return_types():
            # In this case `deepBox` returns `SomeBox` in the first usage, and
            # `StringBox` in the second usage. These types are not the same!
            # However this is valid because the return *shapes* are compatible.
            assert_valid(
                """
                {
                  someBox {
                      ... on SomeBox {
                      deepBox {
                        unrelatedField
                      }
                    }
                    ... on StringBox {
                      deepBox {
                        unrelatedField
                      }
                    }
                  }
                }
                """,
                schema=schema,
            )

        def disallows_differing_return_types_despite_no_overlap():
            assert_errors(
                """
                {
                  someBox {
                    ... on IntBox {
                      scalar
                    }
                    ... on StringBox {
                      scalar
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'scalar' conflict because"
                        " they return conflicting types 'Int' and 'String'."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(5, 23), (8, 23)],
                    }
                ],
                schema,
            )

        def reports_correctly_when_a_non_exclusive_follows_an_exclusive():
            assert_errors(
                """
                {
                  someBox {
                    ... on IntBox {
                      deepBox {
                        ...X
                      }
                    }
                  }
                  someBox {
                    ... on StringBox {
                      deepBox {
                        ...Y
                      }
                    }
                  }
                  memoed: someBox {
                    ... on IntBox {
                      deepBox {
                        ...X
                      }
                    }
                  }
                  memoed: someBox {
                    ... on StringBox {
                      deepBox {
                        ...Y
                      }
                    }
                  }
                  other: someBox {
                    ...X
                  }
                  other: someBox {
                    ...Y
                  }
                }
                fragment X on SomeBox {
                  scalar
                }
                fragment Y on SomeBox {
                  scalar: unrelatedField
                }
                """,
                [
                    {
                        "message": "Fields 'other' conflict because"
                        " subfields 'scalar' conflict because"
                        " 'scalar' and 'unrelatedField' are different fields."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(31, 19), (39, 19), (34, 19), (42, 19)],
                        "path": None,
                    }
                ],
                schema,
            )

        def disallows_differing_return_type_nullability_despite_no_overlap():
            assert_errors(
                """
                {
                  someBox {
                    ... on NonNullStringBox1 {
                      scalar
                    }
                    ... on StringBox {
                      scalar
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'scalar' conflict because"
                        " they return conflicting types 'String!' and 'String'."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(5, 23), (8, 23)],
                    }
                ],
                schema,
            )

        def disallows_differing_return_type_list_despite_no_overlap_1():
            assert_errors(
                """
                {
                  someBox {
                    ... on IntBox {
                      box: listStringBox {
                        scalar
                      }
                    }
                    ... on StringBox {
                      box: stringBox {
                        scalar
                      }
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'box' conflict because they return"
                        " conflicting types '[StringBox]' and 'StringBox'."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(5, 23), (10, 23)],
                    }
                ],
                schema,
            )

            assert_errors(
                """
                {
                  someBox {
                    ... on IntBox {
                      box: stringBox {
                        scalar
                      }
                    }
                    ... on StringBox {
                      box: listStringBox {
                        scalar
                      }
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'box' conflict because they return"
                        " conflicting types 'StringBox' and '[StringBox]'."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(5, 23), (10, 23)],
                    }
                ],
                schema,
            )

        def disallows_differing_subfields():
            assert_errors(
                """
                {
                  someBox {
                    ... on IntBox {
                      box: stringBox {
                        val: scalar
                        val: unrelatedField
                      }
                    }
                    ... on StringBox {
                      box: stringBox {
                        val: scalar
                      }
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'val' conflict because"
                        " 'scalar' and 'unrelatedField' are different fields."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(6, 25), (7, 25)],
                    }
                ],
                schema,
            )

        def disallows_differing_deep_return_types_despite_no_overlap():
            assert_errors(
                """
                {
                  someBox {
                    ... on IntBox {
                      box: stringBox {
                        scalar
                      }
                    }
                    ... on StringBox {
                      box: intBox {
                        scalar
                      }
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'box' conflict"
                        " because subfields 'scalar' conflict"
                        " because they return conflicting types 'String' and 'Int'."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [(5, 23), (6, 25), (10, 23), (11, 25)],
                        "path": None,
                    }
                ],
                schema,
            )

        def allows_non_conflicting_overlapping_types():
            assert_valid(
                """
                {
                  someBox {
                    ... on IntBox {
                      scalar: unrelatedField
                    }
                    ... on StringBox {
                      scalar
                    }
                  }
                }
                """,
                schema=schema,
            )

        def same_wrapped_scalar_return_types():
            assert_valid(
                """
                {
                  someBox {
                    ...on NonNullStringBox1 {
                      scalar
                    }
                    ...on NonNullStringBox2 {
                      scalar
                    }
                  }
                }
                """,
                schema=schema,
            )

        def allows_inline_fragments_without_type_condition():
            assert_valid(
                """
                {
                  a
                  ... {
                    a
                  }
                }
                """,
                schema=schema,
            )

        def compares_deep_types_including_list():
            assert_errors(
                """
                {
                  connection {
                    ...edgeID
                    edges {
                      node {
                        id: name
                      }
                    }
                  }
                }

                fragment edgeID on Connection {
                  edges {
                    node {
                      id
                    }
                  }
                }
                """,
                [
                    {
                        "message": "Fields 'edges' conflict"
                        " because subfields 'node' conflict"
                        " because subfields 'id' conflict"
                        " because 'name' and 'id' are different fields."
                        " Use different aliases on the fields"
                        " to fetch both if this was intentional.",
                        "locations": [
                            (5, 21),
                            (6, 23),
                            (7, 25),
                            (14, 19),
                            (15, 21),
                            (16, 23),
                        ],
                        "path": None,
                    }
                ],
                schema,
            )

        def ignores_unknown_types():
            assert_valid(
                """
                {
                  someBox {
                    ...on UnknownType {
                      scalar
                    }
                    ...on NonNullStringBox2 {
                      scalar
                    }
                  }
                }
                """,
                schema=schema,
            )

        def works_for_field_names_that_are_js_keywords():
            schema_with_keywords = build_schema(
                """
                type Foo {
                  constructor: String
                }

                type Query {
                  foo: Foo
                }
                """
            )

            assert_valid(
                """
                {
                  foo {
                    constructor
                  }
                }
                """,
                schema=schema_with_keywords,
            )

        def works_for_field_names_that_are_python_keywords():
            schema_with_keywords = build_schema(
                """
                type Foo {
                  class: String
                }

                type Query {
                  foo: Foo
                }
                """
            )

            assert_valid(
                """
                {
                  foo {
                    class
                  }
                }
                """,
                schema=schema_with_keywords,
            )

    def does_not_infinite_loop_on_recursive_fragments():
        assert_valid(
            """
            {
              ...fragA
            }

            fragment fragA on Human { name, relatives { name, ...fragA } }
            """
        )

    def does_not_infinite_loop_on_immediately_recursive_fragments():
        assert_valid(
            """
            {
              ...fragA
            }

            fragment fragA on Human { name, ...fragA }
            """
        )

    def does_not_infinite_loop_on_recursive_fragment_with_field_named_after_fragment():
        assert_valid(
            """
            {
              ...fragA
              fragA
            }

            fragment fragA on Query { ...fragA }
            """
        )

    def finds_invalid_cases_even_with_field_named_after_fragment():
        assert_errors(
            """
            {
              fragA
              ...fragA
            }

            fragment fragA on Type {
              fragA: b
            }
            """,
            [
                {
                    "message": "Fields 'fragA' conflict"
                    " because 'fragA' and 'b' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(3, 15), (8, 15)],
                }
            ],
        )

    def does_not_infinite_loop_on_transitively_recursive_fragments():
        assert_valid(
            """
            {
              ...fragA
              fragB
            }

            fragment fragA on Human { name, ...fragB }
            fragment fragB on Human { name, ...fragC }
            fragment fragC on Human { name, ...fragA }
            """
        )

    def finds_invalid_case_even_with_immediately_recursive_fragment():
        assert_errors(
            """
            fragment sameAliasesWithDifferentFieldTargets on Dog {
              ...sameAliasesWithDifferentFieldTargets
              fido: name
              fido: nickname
            }
            """,
            [
                {
                    "message": "Fields 'fido' conflict"
                    " because 'name' and 'nickname' are different fields."
                    " Use different aliases on the fields"
                    " to fetch both if this was intentional.",
                    "locations": [(4, 15), (5, 15)],
                }
            ],
        )



from pytest import mark

from graphql.language import Source, Lexer, TokenKind
from graphql.language.block_string import (
    print_block_string,
    is_printable_as_block_string,
)

from ..utils import dedent, gen_fuzz_strings


def lex_value(s: str) -> str:
    lexer = Lexer(Source(s))
    value = lexer.advance().value
    assert isinstance(value, str)
    assert lexer.advance().kind == TokenKind.EOF, "Expected EOF"
    return value


def assert_printable_block_string(test_value: str, minimize: bool = False) -> None:
    block_string = print_block_string(test_value, minimize=minimize)
    printed_value = lex_value(block_string)
    assert test_value == printed_value, dedent(
        f"""
        Expected lexValue({block_string!r})
          to equal {test_value!r}
          but got  {printed_value!r}
        """
    )


def assert_non_printable_block_string(test_value: str) -> None:
    block_string = print_block_string(test_value)
    printed_value = lex_value(block_string)
    assert test_value != printed_value, dedent(
        f"""
        Expected lexValue({block_string!r})
          to not equal {test_value!r}
        """
    )


def describe_print_block_string():
    @mark.slow
    @mark.timeout(20)
    def correctly_print_random_strings():
        # Testing with length >7 is taking exponentially more time. However, it is
        # highly recommended testing with increased limit if you make any change.
        for fuzz_str in gen_fuzz_strings(allowed_chars='\n\t "a\\', max_length=7):
            if not is_printable_as_block_string(fuzz_str):
                assert_non_printable_block_string(fuzz_str)
                continue

            assert_printable_block_string(fuzz_str)
            assert_printable_block_string(fuzz_str, minimize=True)



from . import gen_fuzz_strings


def describe_gen_fuzz_strings():
    def always_provide_empty_string():
        assert list(gen_fuzz_strings(allowed_chars="", max_length=0)) == [""]
        assert list(gen_fuzz_strings(allowed_chars="", max_length=1)) == [""]
        assert list(gen_fuzz_strings(allowed_chars="a", max_length=0)) == [""]

    def generate_strings_with_single_character():
        assert list(gen_fuzz_strings(allowed_chars="a", max_length=1)) == ["", "a"]
        assert list(gen_fuzz_strings(allowed_chars="abc", max_length=1)) == [
            "",
            "a",
            "b",
            "c",
        ]

    def generate_strings_with_multiple_character():
        assert list(gen_fuzz_strings(allowed_chars="a", max_length=2)) == [
            "",
            "a",
            "aa",
        ]

        assert list(gen_fuzz_strings(allowed_chars="abc", max_length=2)) == [
            "",
            "a",
            "b",
            "c",
            "aa",
            "ab",
            "ac",
            "ba",
            "bb",
            "bc",
            "ca",
            "cb",
            "cc",
        ]

    def generate_strings_longer_than_possible_number_of_characters():
        assert list(gen_fuzz_strings(allowed_chars="ab", max_length=3)) == [
            "",
            "a",
            "b",
            "aa",
            "ab",
            "ba",
            "bb",
            "aaa",
            "aab",
            "aba",
            "abb",
            "baa",
            "bab",
            "bba",
            "bbb",
        ]



from typing import Any

from ...error import GraphQLError
from ...language import FieldNode
from ...type import get_named_type, is_leaf_type
from . import ValidationRule

__all__ = ["ScalarLeafsRule"]


class ScalarLeafsRule(ValidationRule):
    """Scalar leafs

    A GraphQL document is valid only if all leaf fields (fields without sub selections)
    are of scalar or enum types.
    """

    def enter_field(self, node: FieldNode, *_args: Any) -> None:
        type_ = self.context.get_type()
        if type_:
            selection_set = node.selection_set
            if is_leaf_type(get_named_type(type_)):
                if selection_set:
                    field_name = node.name.value
                    self.report_error(
                        GraphQLError(
                            f"Field '{field_name}' must not have a selection"
                            f" since type '{type_}' has no subfields.",
                            selection_set,
                        )
                    )
            elif not selection_set:
                field_name = node.name.value
                self.report_error(
                    GraphQLError(
                        f"Field '{field_name}' of type '{type_}'"
                        " must have a selection of subfields."
                        f" Did you mean '{field_name} {{ ... }}'?",
                        node,
                    )
                )



from typing import Any, List

from ...error import GraphQLError
from ...language import (
    FragmentDefinitionNode,
    OperationDefinitionNode,
    VisitorAction,
    SKIP,
)
from . import ASTValidationContext, ASTValidationRule

__all__ = ["NoUnusedFragmentsRule"]


class NoUnusedFragmentsRule(ASTValidationRule):
    """No unused fragments

    A GraphQL document is only valid if all fragment definitions are spread within
    operations, or spread within other fragments spread within operations.

    See https://spec.graphql.org/draft/#sec-Fragments-Must-Be-Used
    """

    def __init__(self, context: ASTValidationContext):
        super().__init__(context)
        self.operation_defs: List[OperationDefinitionNode] = []
        self.fragment_defs: List[FragmentDefinitionNode] = []

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_args: Any
    ) -> VisitorAction:
        self.operation_defs.append(node)
        return SKIP

    def enter_fragment_definition(
        self, node: FragmentDefinitionNode, *_args: Any
    ) -> VisitorAction:
        self.fragment_defs.append(node)
        return SKIP

    def leave_document(self, *_args: Any) -> None:
        fragment_names_used = set()
        get_fragments = self.context.get_recursively_referenced_fragments
        for operation in self.operation_defs:
            for fragment in get_fragments(operation):
                fragment_names_used.add(fragment.name.value)

        for fragment_def in self.fragment_defs:
            frag_name = fragment_def.name.value
            if frag_name not in fragment_names_used:
                self.report_error(
                    GraphQLError(f"Fragment '{frag_name}' is never used.", fragment_def)
                )



from graphql.utilities import build_schema, print_schema, lexicographic_sort_schema

from ..utils import dedent


def sort_sdl(sdl: str) -> str:
    schema = build_schema(sdl)
    return print_schema(lexicographic_sort_schema(schema))


def describe_lexicographic_sort_schema():
    def sort_fields():
        sorted_sdl = sort_sdl(
            """
            input Bar {
              barB: String!
              barA: String
              barC: [String]
            }

            interface FooInterface {
              fooB: String!
              fooA: String
              fooC: [String]
            }

            type FooType implements FooInterface {
              fooC: [String]
              fooA: String
              fooB: String!
            }

            type Query {
              dummy(arg: Bar): FooType
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            input Bar {
              barA: String
              barB: String!
              barC: [String]
            }

            interface FooInterface {
              fooA: String
              fooB: String!
              fooC: [String]
            }

            type FooType implements FooInterface {
              fooA: String
              fooB: String!
              fooC: [String]
            }

            type Query {
              dummy(arg: Bar): FooType
            }
            """
        )

    def sort_implemented_interfaces():
        sorted_sdl = sort_sdl(
            """
            interface FooA {
              dummy: String
            }

            interface FooB {
              dummy: String
            }

            interface FooC implements FooB & FooA {
              dummy: String
            }

            type Query implements FooB & FooA & FooC {
              dummy: String
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            interface FooA {
              dummy: String
            }

            interface FooB {
              dummy: String
            }

            interface FooC implements FooA & FooB {
              dummy: String
            }

            type Query implements FooA & FooB & FooC {
              dummy: String
            }
            """
        )

    def sort_types_in_union():
        sorted_sdl = sort_sdl(
            """
            type FooA {
              dummy: String
            }

            type FooB {
              dummy: String
            }

            type FooC {
              dummy: String
            }

            union FooUnion = FooB | FooA | FooC

            type Query {
              dummy: FooUnion
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            type FooA {
              dummy: String
            }

            type FooB {
              dummy: String
            }

            type FooC {
              dummy: String
            }

            union FooUnion = FooA | FooB | FooC

            type Query {
              dummy: FooUnion
            }
            """
        )

    def sort_enum_types():
        sorted_sdl = sort_sdl(
            """
            enum Foo {
              B
              C
              A
            }

            type Query {
              dummy: Foo
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            enum Foo {
              A
              B
              C
            }

            type Query {
              dummy: Foo
            }
            """
        )

    def sort_field_arguments():
        sorted_sdl = sort_sdl(
            """
            type Query {
              dummy(argB: Int!, argA: String, argC: [Float]): ID
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            type Query {
              dummy(argA: String, argB: Int!, argC: [Float]): ID
            }
            """
        )

    def sort_types():
        sorted_sdl = sort_sdl(
            """
            type Query {
              dummy(arg1: FooF, arg2: FooA, arg3: FooG): FooD
            }

            type FooC implements FooE {
              dummy: String
            }

            enum FooG {
              enumValue
            }

            scalar FooA

            input FooF {
              dummy: String
            }

            union FooD = FooC | FooB

            interface FooE {
              dummy: String
            }

            type FooB {
              dummy: String
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            scalar FooA

            type FooB {
              dummy: String
            }

            type FooC implements FooE {
              dummy: String
            }

            union FooD = FooB | FooC

            interface FooE {
              dummy: String
            }

            input FooF {
              dummy: String
            }

            enum FooG {
              enumValue
            }

            type Query {
              dummy(arg1: FooF, arg2: FooA, arg3: FooG): FooD
            }
            """
        )

    def sort_directive_arguments():
        sorted_sdl = sort_sdl(
            """
            directive @test(argC: Float, argA: String, argB: Int) on FIELD

            type Query {
              dummy: String
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            directive @test(argA: String, argB: Int, argC: Float) on FIELD

            type Query {
              dummy: String
            }
            """
        )

    def sort_directive_locations():
        sorted_sdl = sort_sdl(
            """
            directive @test(argC: Float, argA: String, argB: Int) on UNION | FIELD | ENUM

            type Query {
              dummy: String
            }
            """  # noqa: E501
        )

        assert sorted_sdl == dedent(
            """
            directive @test(argA: String, argB: Int, argC: Float) on ENUM | FIELD | UNION

            type Query {
              dummy: String
            }
            """  # noqa: E501
        )

    def sort_directives():
        sorted_sdl = sort_sdl(
            """
            directive @fooC on FIELD

            directive @fooB on UNION

            directive @fooA on ENUM

            type Query {
              dummy: String
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            directive @fooA on ENUM

            directive @fooB on UNION

            directive @fooC on FIELD

            type Query {
              dummy: String
            }
            """
        )

    def sort_recursive_types():
        sorted_sdl = sort_sdl(
            """
            interface FooC {
              fooB: FooB
              fooA: FooA
              fooC: FooC
            }

            type FooB implements FooC {
              fooB: FooB
              fooA: FooA
            }

            type FooA implements FooC {
              fooB: FooB
              fooA: FooA
            }

            type Query {
              fooC: FooC
              fooB: FooB
              fooA: FooA
            }
            """
        )

        assert sorted_sdl == dedent(
            """
            type FooA implements FooC {
              fooA: FooA
              fooB: FooB
            }

            type FooB implements FooC {
              fooA: FooA
              fooB: FooB
            }

            interface FooC {
              fooA: FooA
              fooB: FooB
              fooC: FooC
            }

            type Query {
              fooA: FooA
              fooB: FooB
              fooC: FooC
            }
            """
        )



from enum import Enum

__all__ = ["DirectiveLocation"]


class DirectiveLocation(Enum):
    """The enum type representing the directive location values."""

    # Request Definitions
    QUERY = "query"
    MUTATION = "mutation"
    SUBSCRIPTION = "subscription"
    FIELD = "field"
    FRAGMENT_DEFINITION = "fragment definition"
    FRAGMENT_SPREAD = "fragment spread"
    VARIABLE_DEFINITION = "variable definition"
    INLINE_FRAGMENT = "inline fragment"

    # Type System Definitions
    SCHEMA = "schema"
    SCALAR = "scalar"
    OBJECT = "object"
    FIELD_DEFINITION = "field definition"
    ARGUMENT_DEFINITION = "argument definition"
    INTERFACE = "interface"
    UNION = "union"
    ENUM = "enum"
    ENUM_VALUE = "enum value"
    INPUT_OBJECT = "input object"
    INPUT_FIELD_DEFINITION = "input field definition"



from functools import partial

from graphql.utilities import build_schema
from graphql.validation.rules.unique_type_names import UniqueTypeNamesRule

from .harness import assert_sdl_validation_errors

assert_errors = partial(assert_sdl_validation_errors, UniqueTypeNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_type_names():
    def no_types():
        assert_valid(
            """
            directive @test on SCHEMA
            """
        )

    def one_type():
        assert_valid(
            """
            type Foo
            """
        )

    def many_types():
        assert_valid(
            """
            type Foo
            type Bar
            type Baz
            """
        )

    def type_and_non_type_definitions_named_the_same():
        assert_valid(
            """
            query Foo { __typename }
            fragment Foo on Query { __typename }
            directive @Foo on SCHEMA

            type Foo
            """
        )

    def types_named_the_same():
        assert_errors(
            """
            type Foo

            scalar Foo
            type Foo
            interface Foo
            union Foo
            enum Foo
            input Foo
            """,
            [
                {
                    "message": "There can be only one type named 'Foo'.",
                    "locations": [(2, 18), (4, 20)],
                },
                {
                    "message": "There can be only one type named 'Foo'.",
                    "locations": [(2, 18), (5, 18)],
                },
                {
                    "message": "There can be only one type named 'Foo'.",
                    "locations": [(2, 18), (6, 23)],
                },
                {
                    "message": "There can be only one type named 'Foo'.",
                    "locations": [(2, 18), (7, 19)],
                },
                {
                    "message": "There can be only one type named 'Foo'.",
                    "locations": [(2, 18), (8, 18)],
                },
                {
                    "message": "There can be only one type named 'Foo'.",
                    "locations": [(2, 18), (9, 19)],
                },
            ],
        )

    def adding_new_type_to_existing_schema():
        schema = build_schema("type Foo")

        assert_valid("type Bar", schema=schema)

    def adding_new_type_to_existing_schema_with_same_named_directive():
        schema = build_schema("directive @Foo on SCHEMA")

        assert_valid("type Foo", schema=schema)

    def adding_conflicting_types_to_existing_schema():
        schema = build_schema("type Foo")
        sdl = """
            scalar Foo
            type Foo
            interface Foo
            union Foo
            enum Foo
            input Foo
            """

        assert_errors(
            sdl,
            [
                {
                    "message": "Type 'Foo' already exists in the schema."
                    " It cannot also be defined in this type definition.",
                    "locations": [(2, 20)],
                },
                {
                    "message": "Type 'Foo' already exists in the schema."
                    " It cannot also be defined in this type definition.",
                    "locations": [(3, 18)],
                },
                {
                    "message": "Type 'Foo' already exists in the schema."
                    " It cannot also be defined in this type definition.",
                    "locations": [(4, 23)],
                },
                {
                    "message": "Type 'Foo' already exists in the schema."
                    " It cannot also be defined in this type definition.",
                    "locations": [(5, 19)],
                },
                {
                    "message": "Type 'Foo' already exists in the schema."
                    " It cannot also be defined in this type definition.",
                    "locations": [(6, 18)],
                },
                {
                    "message": "Type 'Foo' already exists in the schema."
                    " It cannot also be defined in this type definition.",
                    "locations": [(7, 19)],
                },
            ],
            schema,
        )



from functools import partial

from graphql.utilities import build_schema
from graphql.validation.rules.unique_enum_value_names import UniqueEnumValueNamesRule

from .harness import assert_sdl_validation_errors

assert_errors = partial(assert_sdl_validation_errors, UniqueEnumValueNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_field_definition_names():
    def no_values():
        assert_valid(
            """
            enum SomeEnum
            """
        )

    def one_value():
        assert_valid(
            """
            enum SomeEnum {
              FOO
            }
            """
        )

    def multiple_values():
        assert_valid(
            """
            enum SomeEnum {
              FOO
              BAR
            }
            """
        )

    def duplicate_values_inside_the_same_enum_definition():
        assert_errors(
            """
            enum SomeEnum {
              FOO
              BAR
              FOO
            }
            """,
            [
                {
                    "message": "Enum value 'SomeEnum.FOO' can only be defined once.",
                    "locations": [(3, 15), (5, 15)],
                },
            ],
        )

    def extend_enum_with_new_value():
        assert_valid(
            """
            enum SomeEnum {
              FOO
            }
            extend enum SomeEnum {
              BAR
            }
            extend enum SomeEnum {
              BAZ
            }
            """
        )

    def extend_enum_with_duplicate_value():
        assert_errors(
            """
            extend enum SomeEnum {
              FOO
            }
            enum SomeEnum {
              FOO
            }
            """,
            [
                {
                    "message": "Enum value 'SomeEnum.FOO' can only be defined once.",
                    "locations": [(3, 15), (6, 15)],
                },
            ],
        )

    def duplicate_value_inside_extension():
        assert_errors(
            """
            enum SomeEnum
            extend enum SomeEnum {
              FOO
              BAR
              FOO
            }
            """,
            [
                {
                    "message": "Enum value 'SomeEnum.FOO' can only be defined once.",
                    "locations": [(4, 15), (6, 15)],
                },
            ],
        )

    def duplicate_value_inside_different_extension():
        assert_errors(
            """
            enum SomeEnum
            extend enum SomeEnum {
              FOO
            }
            extend enum SomeEnum {
              FOO
            }
            """,
            [
                {
                    "message": "Enum value 'SomeEnum.FOO' can only be defined once.",
                    "locations": [(4, 15), (7, 15)],
                },
            ],
        )

    def adding_new_value_to_the_enum_inside_existing_schema():
        schema = build_schema("enum SomeEnum")
        sdl = """
          extend enum SomeEnum {
              FOO
          }
          """

        assert_valid(sdl, schema=schema)

    def adding_conflicting_value_to_existing_schema_twice():
        schema = build_schema(
            """
            enum SomeEnum {
              FOO
            }
            """
        )
        sdl = """
            extend enum SomeEnum {
              FOO
            }
            extend enum SomeEnum {
              FOO
            }
            """

        assert_errors(
            sdl,
            [
                {
                    "message": "Enum value 'SomeEnum.FOO' already exists in the schema."
                    " It cannot also be defined in this type extension.",
                    "locations": [(3, 15)],
                },
                {
                    "message": "Enum value 'SomeEnum.FOO' already exists in the schema."
                    " It cannot also be defined in this type extension.",
                    "locations": [(6, 15)],
                },
            ],
            schema,
        )

    def adding_enum_values_to_existing_schema_twice():
        schema = build_schema("enum SomeEnum")
        sdl = """
            extend enum SomeEnum {
              FOO
            }
            extend enum SomeEnum {
              FOO
            }
            """

        assert_errors(
            sdl,
            [
                {
                    "message": "Enum value 'SomeEnum.FOO' can only be defined once.",
                    "locations": [(3, 15), (6, 15)],
                },
            ],
            schema,
        )



from functools import partial

from graphql.validation import UniqueVariableNamesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, UniqueVariableNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_variable_names():
    def unique_variable_names():
        assert_valid(
            """
            query A($x: Int, $y: String) { __typename }
            query B($x: String, $y: Int) { __typename }
            """
        )

    def duplicate_variable_names():
        assert_errors(
            """
            query A($x: Int, $x: Int, $x: String) { __typename }
            query B($x: String, $x: Int) { __typename }
            query C($x: Int, $x: Int) { __typename }
            """,
            [
                {
                    "message": "There can be only one variable named '$x'.",
                    "locations": [(2, 22), (2, 31), (2, 40)],
                },
                {
                    "message": "There can be only one variable named '$x'.",
                    "locations": [(3, 22), (3, 34)],
                },
                {
                    "message": "There can be only one variable named '$x'.",
                    "locations": [(4, 22), (4, 31)],
                },
            ],
        )



from copy import deepcopy

from pytest import raises

from graphql.language import FieldNode, NameNode, parse, print_ast

from ..fixtures import kitchen_sink_query  # noqa: F401
from ..utils import dedent


def describe_printer_query_document():
    def prints_minimal_ast():
        ast = FieldNode(name=NameNode(value="foo"))
        assert print_ast(ast) == "foo"

    def produces_helpful_error_messages():
        bad_ast = {"random": "Data"}
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            print_ast(bad_ast)  # type: ignore
        assert str(exc_info.value) == "Not an AST Node: {'random': 'Data'}."
        corrupt_ast = FieldNode(name="random data")
        with raises(TypeError) as exc_info:
            print_ast(corrupt_ast)
        assert str(exc_info.value) == "Invalid AST Node: 'random data'."

    def correctly_prints_query_operation_without_name():
        query_ast_shorthanded = parse("query { id, name }")
        assert print_ast(query_ast_shorthanded) == "{\n  id\n  name\n}"

    def correctly_prints_mutation_operation_without_name():
        mutation_ast = parse("mutation { id, name }")
        assert print_ast(mutation_ast) == "mutation {\n  id\n  name\n}"

    def correctly_prints_query_operation_with_artifacts():
        query_ast_with_artifacts = parse(
            "query ($foo: TestType) @testDirective { id, name }"
        )
        assert print_ast(query_ast_with_artifacts) == dedent(
            """
            query ($foo: TestType) @testDirective {
              id
              name
            }
            """
        )

    def correctly_prints_mutation_operation_with_artifacts():
        mutation_ast_with_artifacts = parse(
            "mutation ($foo: TestType) @testDirective { id, name }"
        )
        assert print_ast(mutation_ast_with_artifacts) == dedent(
            """
            mutation ($foo: TestType) @testDirective {
              id
              name
            }
            """
        )

    def prints_query_with_variable_directives():
        query_ast_with_variable_directive = parse(
            "query ($foo: TestType = {a: 123}" " @testDirective(if: true) @test) { id }"
        )
        assert print_ast(query_ast_with_variable_directive) == dedent(
            """
            query ($foo: TestType = {a: 123} @testDirective(if: true) @test) {
              id
            }
            """
        )

    def keeps_arguments_on_one_line_if_line_has_80_chars_or_less():
        printed = print_ast(parse("{trip(wheelchair:false arriveBy:false){dateTime}}"))

        assert printed == dedent(
            """
            {
              trip(wheelchair: false, arriveBy: false) {
                dateTime
              }
            }
            """
        )

    def puts_arguments_on_multiple_lines_if_line_has_more_than_80_chars():
        printed = print_ast(
            parse(
                "{trip(wheelchair:false arriveBy:false includePlannedCancellations:true"
                " transitDistanceReluctance:2000){dateTime}}"
            )
        )

        assert printed == dedent(
            """
            {
              trip(
                wheelchair: false
                arriveBy: false
                includePlannedCancellations: true
                transitDistanceReluctance: 2000
              ) {
                dateTime
              }
            }
            """
        )

    def legacy_prints_fragment_with_variable_directives():
        query_ast_with_variable_directive = parse(
            "fragment Foo($foo: TestType @test) on TestType @testDirective { id }",
            allow_legacy_fragment_variables=True,
        )
        assert print_ast(query_ast_with_variable_directive) == dedent(
            """
            fragment Foo($foo: TestType @test) on TestType @testDirective {
              id
            }
            """
        )

    def legacy_correctly_prints_fragment_defined_variables():
        source = """
            fragment Foo($a: ComplexType, $b: Boolean = false) on TestType {
              id
            }
            """
        fragment_with_variable = parse(source, allow_legacy_fragment_variables=True)
        assert print_ast(fragment_with_variable) == dedent(source)

    def prints_kitchen_sink_without_altering_ast(kitchen_sink_query):  # noqa: F811
        ast = parse(kitchen_sink_query, no_location=True)

        ast_before_print_call = deepcopy(ast)
        printed = print_ast(ast)
        printed_ast = parse(printed, no_location=True)
        assert printed_ast == ast
        assert deepcopy(ast) == ast_before_print_call

        assert printed == dedent(
            r'''
            query queryName($foo: ComplexType, $site: Site = MOBILE) @onQuery {
              whoever123is: node(id: [123, 456]) {
                id
                ... on User @onInlineFragment {
                  field2 {
                    id
                    alias: field1(first: 10, after: $foo) @include(if: $foo) {
                      id
                      ...frag @onFragmentSpread
                    }
                  }
                }
                ... @skip(unless: $foo) {
                  id
                }
                ... {
                  id
                }
              }
            }

            mutation likeStory @onMutation {
              like(story: 123) @onField {
                story {
                  id @onField
                }
              }
            }

            subscription StoryLikeSubscription($input: StoryLikeSubscribeInput @onVariableDefinition) @onSubscription {
              storyLikeSubscribe(input: $input) {
                story {
                  likers {
                    count
                  }
                  likeSentence {
                    text
                  }
                }
              }
            }

            fragment frag on Friend @onFragmentDefinition {
              foo(
                size: $size
                bar: $b
                obj: {key: "value", block: """
                block string uses \"""
                """}
              )
            }

            {
              unnamed(truthy: true, falsy: false, nullish: null)
              query
            }

            {
              __typename
            }
            '''  # noqa: E501
        )



from collections import ChainMap
from typing import Any

from graphql.error import GraphQLError
from graphql.execution import execute_sync, ExecutionResult
from graphql.language import parse, SourceLocation
from graphql.type import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)


def describe_execute_resolve_function():
    def _test_schema(test_field: GraphQLField) -> GraphQLSchema:
        return GraphQLSchema(GraphQLObjectType("Query", {"test": test_field}))

    def default_function_accesses_attributes():
        class RootValue:
            test = "testValue"

        assert execute_sync(
            schema=_test_schema(GraphQLField(GraphQLString)),
            document=parse("{ test }"),
            root_value=RootValue(),
        ) == (
            {"test": "testValue"},
            None,
        )

    def default_function_accesses_keys_of_dict():
        root_value = {"test": "testValue"}

        assert execute_sync(
            schema=_test_schema(GraphQLField(GraphQLString)),
            document=parse("{ test }"),
            root_value=root_value,
        ) == ({"test": "testValue"}, None)

    def default_function_accesses_keys_of_chain_map():
        # use a mapping that is not a subclass of dict
        root_value = ChainMap({"test": "testValue"})

        assert execute_sync(
            schema=_test_schema(GraphQLField(GraphQLString)),
            document=parse("{ test }"),
            root_value=root_value,
        ) == ({"test": "testValue"}, None)

    def default_function_calls_methods():
        class RootValue:
            _secret = "secretValue"

            def test(self, _info):
                return self._secret

        assert execute_sync(
            schema=_test_schema(GraphQLField(GraphQLString)),
            document=parse("{ test }"),
            root_value=RootValue(),
        ) == (
            {"test": "secretValue"},
            None,
        )

    def default_function_passes_args_and_context():
        class Adder:
            _num: int

            def __init__(self, num):
                self._num = num

            def test(self, info, addend1: int):
                return self._num + addend1 + info.context.addend2

        root_value = Adder(700)

        schema = _test_schema(
            GraphQLField(GraphQLInt, args={"addend1": GraphQLArgument(GraphQLInt)})
        )

        class ContextValue:
            addend2 = 9

        context_value = ContextValue()
        document = parse("{ test(addend1: 80) }")

        assert execute_sync(
            schema=schema,
            document=document,
            root_value=root_value,
            context_value=context_value,
        ) == (
            {"test": 789},
            None,
        )

    def uses_provided_resolve_function():
        schema = _test_schema(
            GraphQLField(
                GraphQLString,
                args={
                    "aStr": GraphQLArgument(GraphQLString),
                    "aInt": GraphQLArgument(GraphQLInt),
                },
                resolve=lambda source, info, **args: repr([source, args]),
            )
        )

        def execute_query(query: str, root_value: Any = None) -> ExecutionResult:
            document = parse(query)
            return execute_sync(
                schema=schema,
                document=document,
                root_value=root_value,
            )

        assert execute_query("{ test }") == ({"test": "[None, {}]"}, None)

        assert execute_query("{ test }", "Source!") == (
            {"test": "['Source!', {}]"},
            None,
        )

        assert execute_query('{ test(aStr: "String!") }', "Source!") == (
            {"test": "['Source!', {'aStr': 'String!'}]"},
            None,
        )

        assert execute_query('{ test(aInt: -123, aStr: "String!") }', "Source!") == (
            {"test": "['Source!', {'aStr': 'String!', 'aInt': -123}]"},
            None,
        )

    def transforms_arguments_using_out_names():
        # This is an extension of GraphQL.js.
        schema = _test_schema(
            GraphQLField(
                GraphQLString,
                args={
                    "aStr": GraphQLArgument(GraphQLString, out_name="a_str"),
                    "aInt": GraphQLArgument(GraphQLInt, out_name="a_int"),
                },
                resolve=lambda source, info, **args: repr([source, args]),
            )
        )

        def execute_query(query: str, root_value: Any = None) -> ExecutionResult:
            document = parse(query)
            return execute_sync(schema=schema, document=document, root_value=root_value)

        assert execute_query("{ test }") == ({"test": "[None, {}]"}, None)

        assert execute_query("{ test }", "Source!") == (
            {"test": "['Source!', {}]"},
            None,
        )

        assert execute_query('{ test(aStr: "String!") }', "Source!") == (
            {"test": "['Source!', {'a_str': 'String!'}]"},
            None,
        )

        assert execute_query('{ test(aInt: -123, aStr: "String!") }', "Source!") == (
            {"test": "['Source!', {'a_str': 'String!', 'a_int': -123}]"},
            None,
        )

    def transforms_arguments_with_inputs_using_out_names():
        # This is an extension of GraphQL.js.
        TestInputObject = GraphQLInputObjectType(
            "TestInputObjectType",
            lambda: {
                "inputOne": GraphQLInputField(GraphQLString, out_name="input_one"),
                "inputRecursive": GraphQLInputField(
                    TestInputObject, out_name="input_recursive"
                ),
            },
        )

        schema = _test_schema(
            GraphQLField(
                GraphQLString,
                args={"aInput": GraphQLArgument(TestInputObject, out_name="a_input")},
                resolve=lambda source, info, **args: repr([source, args]),
            )
        )

        def execute_query(query: str, root_value: Any = None) -> ExecutionResult:
            document = parse(query)
            return execute_sync(schema=schema, document=document, root_value=root_value)

        assert execute_query("{ test }") == ({"test": "[None, {}]"}, None)

        assert execute_query('{ test(aInput: {inputOne: "String!"}) }', "Source!") == (
            {"test": "['Source!', {'a_input': {'input_one': 'String!'}}]"},
            None,
        )

        assert execute_query(
            '{ test(aInput: {inputRecursive: {inputOne: "SourceRecursive!"}}) }',
            "Source!",
        ) == (
            {
                "test": "['Source!',"
                " {'a_input': {'input_recursive': {'input_one': 'SourceRecursive!'}}}]"
            },
            None,
        )

    def pass_error_from_resolver_wrapped_as_located_graphql_error():
        def resolve(_obj, _info):
            raise ValueError("Some error")

        schema = _test_schema(GraphQLField(GraphQLString, resolve=resolve))
        result = execute_sync(schema, parse("{ test }"))

        assert result == (
            {"test": None},
            [{"message": "Some error", "locations": [(1, 3)], "path": ["test"]}],
        )

        assert result.errors is not None
        error = result.errors[0]
        assert isinstance(error, GraphQLError)
        assert str(error) == "Some error\n\nGraphQL request:1:3\n1 | { test }\n  |   ^"
        assert error.positions == [2]
        locations = error.locations
        assert locations == [(1, 3)]
        location = locations[0]
        assert isinstance(location, SourceLocation)
        assert location == SourceLocation(1, 3)
        original_error = error.original_error
        assert isinstance(original_error, ValueError)
        assert str(original_error) == "Some error"



from typing import Optional

from pytest import mark

from graphql.error import GraphQLSyntaxError
from graphql.language import Lexer, Source, TokenKind
from graphql.utilities import strip_ignored_characters

from ..utils import dedent, gen_fuzz_strings


def lex_value(s: str) -> Optional[str]:
    lexer = Lexer(Source(s))
    value = lexer.advance().value
    assert lexer.advance().kind == TokenKind.EOF, "Expected EOF"
    return value


def describe_strip_ignored_characters():
    @mark.slow
    @mark.timeout(20)
    def strips_ignored_characters_inside_random_block_strings():
        # Testing with length >7 is taking exponentially more time. However it is
        # highly recommended to test with increased limit if you make any change.
        for fuzz_str in gen_fuzz_strings(allowed_chars='\n\t "a\\', max_length=7):
            test_str = f'"""{fuzz_str}"""'

            try:
                test_value = lex_value(test_str)
            except (AssertionError, GraphQLSyntaxError):
                continue  # skip invalid values

            stripped_value = lex_value(strip_ignored_characters(test_str))

            assert test_value == stripped_value, dedent(
                f"""
                Expected lexValue(stripIgnoredCharacters({test_str!r})
                  to equal {test_value!r}
                  but got {stripped_value!r}
                """
            )



import inspect
from typing import Any
from types import CoroutineType, GeneratorType

__all__ = ["is_awaitable"]

CO_ITERABLE_COROUTINE = inspect.CO_ITERABLE_COROUTINE


def is_awaitable(value: Any) -> bool:
    """Return true if object can be passed to an ``await`` expression.

    Instead of testing if the object is an instance of abc.Awaitable, it checks
    the existence of an `__await__` attribute. This is much faster.
    """
    return (
        # check for coroutine objects
        isinstance(value, CoroutineType)
        # check for old-style generator based coroutine objects
        or isinstance(value, GeneratorType)
        and bool(value.gi_code.co_flags & CO_ITERABLE_COROUTINE)
        # check for other awaitables (e.g. futures)
        or hasattr(value, "__await__")
    )



from functools import partial

from graphql.validation import NoUndefinedVariablesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, NoUndefinedVariablesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_no_undefined_variables():
    def all_variables_defined():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              field(a: $a, b: $b, c: $c)
            }
            """
        )

    def all_variables_deeply_defined():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              field(a: $a) {
                field(b: $b) {
                  field(c: $c)
                }
              }
            }
            """
        )

    def all_variables_deeply_in_inline_fragments_defined():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              ... on Type {
                field(a: $a) {
                  field(b: $b) {
                    ... on Type {
                      field(c: $c)
                    }
                  }
                }
              }
            }
            """
        )

    def all_variables_in_fragments_deeply_defined():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragB
              }
            }
            fragment FragB on Type {
              field(b: $b) {
                ...FragC
              }
            }
            fragment FragC on Type {
              field(c: $c)
            }
            """
        )

    def variable_within_single_fragment_defined_in_multiple_operations():
        assert_valid(
            """
            query Foo($a: String) {
              ...FragA
            }
            query Bar($a: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a)
            }
            """
        )

    def variable_within_fragments_defined_in_operations():
        assert_valid(
            """
            query Foo($a: String) {
              ...FragA
            }
            query Bar($b: String) {
              ...FragB
            }
            fragment FragA on Type {
              field(a: $a)
            }
            fragment FragB on Type {
              field(b: $b)
            }
            """
        )

    def variable_within_recursive_fragment_defined():
        assert_valid(
            """
            query Foo($a: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragA
              }
            }
            """
        )

    def variable_not_defined():
        assert_errors(
            """
            query Foo($a: String, $b: String, $c: String) {
              field(a: $a, b: $b, c: $c, d: $d)
            }
            """,
            [
                {
                    "message": "Variable '$d' is not defined by operation 'Foo'.",
                    "locations": [(3, 45), (2, 13)],
                },
            ],
        )

    def variable_not_defined_by_unnamed_query():
        assert_errors(
            """
            {
              field(a: $a)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined.",
                    "locations": [(3, 24), (2, 13)],
                },
            ],
        )

    def multiple_variables_not_defined():
        assert_errors(
            """
            query Foo($b: String) {
              field(a: $a, b: $b, c: $c)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined by operation 'Foo'.",
                    "locations": [(3, 24), (2, 13)],
                },
                {
                    "message": "Variable '$c' is not defined by operation 'Foo'.",
                    "locations": [(3, 38), (2, 13)],
                },
            ],
        )

    def variable_in_fragment_not_defined_by_unnamed_query():
        assert_errors(
            """
            {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined.",
                    "locations": [(6, 24), (2, 13)],
                },
            ],
        )

    def variable_in_fragment_not_defined_by_operation():
        assert_errors(
            """
            query Foo($a: String, $b: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragB
              }
            }
            fragment FragB on Type {
              field(b: $b) {
                ...FragC
              }
            }
            fragment FragC on Type {
              field(c: $c)
            }
            """,
            [
                {
                    "message": "Variable '$c' is not defined by operation 'Foo'.",
                    "locations": [(16, 24), (2, 13)],
                },
            ],
        )

    def multiple_variables_in_fragments_not_defined():
        assert_errors(
            """
            query Foo($b: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragB
              }
            }
            fragment FragB on Type {
              field(b: $b) {
                ...FragC
              }
            }
            fragment FragC on Type {
              field(c: $c)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined by operation 'Foo'.",
                    "locations": [(6, 24), (2, 13)],
                },
                {
                    "message": "Variable '$c' is not defined by operation 'Foo'.",
                    "locations": [(16, 24), (2, 13)],
                },
            ],
        )

    def single_variable_in_fragment_not_defined_by_multiple_operations():
        assert_errors(
            """
            query Foo($a: String) {
              ...FragAB
            }
            query Bar($a: String) {
              ...FragAB
            }
            fragment FragAB on Type {
              field(a: $a, b: $b)
            }
            """,
            [
                {
                    "message": "Variable '$b' is not defined by operation 'Foo'.",
                    "locations": [(9, 31), (2, 13)],
                },
                {
                    "message": "Variable '$b' is not defined by operation 'Bar'.",
                    "locations": [(9, 31), (5, 13)],
                },
            ],
        )

    def variables_in_fragment_not_defined_by_multiple_operations():
        assert_errors(
            """
            query Foo($b: String) {
              ...FragAB
            }
            query Bar($a: String) {
              ...FragAB
            }
            fragment FragAB on Type {
              field(a: $a, b: $b)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined by operation 'Foo'.",
                    "locations": [(9, 24), (2, 13)],
                },
                {
                    "message": "Variable '$b' is not defined by operation 'Bar'.",
                    "locations": [(9, 31), (5, 13)],
                },
            ],
        )

    def variable_in_fragment_used_by_other_operation():
        assert_errors(
            """
            query Foo($b: String) {
              ...FragA
            }
            query Bar($a: String) {
              ...FragB
            }
            fragment FragA on Type {
              field(a: $a)
            }
            fragment FragB on Type {
              field(b: $b)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined by operation 'Foo'.",
                    "locations": [(9, 24), (2, 13)],
                },
                {
                    "message": "Variable '$b' is not defined by operation 'Bar'.",
                    "locations": [(12, 24), (5, 13)],
                },
            ],
        )

    def multiple_undefined_variables_produce_multiple_errors():
        assert_errors(
            """
            query Foo($b: String) {
              ...FragAB
            }
            query Bar($a: String) {
              ...FragAB
            }
            fragment FragAB on Type {
              field1(a: $a, b: $b)
              ...FragC
              field3(a: $a, b: $b)
            }
            fragment FragC on Type {
              field2(c: $c)
            }
            """,
            [
                {
                    "message": "Variable '$a' is not defined by operation 'Foo'.",
                    "locations": [(9, 25), (2, 13)],
                },
                {
                    "message": "Variable '$a' is not defined by operation 'Foo'.",
                    "locations": [(11, 25), (2, 13)],
                },
                {
                    "message": "Variable '$c' is not defined by operation 'Foo'.",
                    "locations": [(14, 25), (2, 13)],
                },
                {
                    "message": "Variable '$b' is not defined by operation 'Bar'.",
                    "locations": [(9, 32), (5, 13)],
                },
                {
                    "message": "Variable '$b' is not defined by operation 'Bar'.",
                    "locations": [(11, 32), (5, 13)],
                },
                {
                    "message": "Variable '$c' is not defined by operation 'Bar'.",
                    "locations": [(14, 25), (5, 13)],
                },
            ],
        )



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import KnownTypeNamesRule

from .harness import assert_validation_errors, assert_sdl_validation_errors

assert_errors = partial(assert_validation_errors, KnownTypeNamesRule)

assert_valid = partial(assert_errors, errors=[])

assert_sdl_errors = partial(assert_sdl_validation_errors, KnownTypeNamesRule)

assert_sdl_valid = partial(assert_sdl_errors, errors=[])


def describe_validate_known_type_names():
    def known_type_names_are_valid():
        assert_valid(
            """
            query Foo(
              $var: String
              $required: [Int!]!
              $introspectionType: __EnumValue
            ) {
              user(id: 4) {
                pets { ... on Pet { name }, ...PetFields, ... { name } }
              }
            }

            fragment PetFields on Pet {
              name
            }
            """
        )

    def unknown_type_names_are_invalid():
        assert_errors(
            """
            query Foo($var: [JumbledUpLetters!]!) {
              user(id: 4) {
                name
                pets { ... on Badger { name }, ...PetFields, ... { name } }
              }
            }
            fragment PetFields on Peat {
              name
            }
            """,
            [
                {
                    "message": "Unknown type 'JumbledUpLetters'.",
                    "locations": [(2, 30)],
                },
                {"message": "Unknown type 'Badger'.", "locations": [(5, 31)]},
                {
                    "message": "Unknown type 'Peat'. Did you mean 'Pet' or 'Cat'?",
                    "locations": [(8, 35)],
                },
            ],
        )

    def references_to_standard_scalars_that_are_missing_in_schema():
        schema = build_schema("type Query { foo: String }")
        query = """
            query ($id: ID, $float: Float, $int: Int) {
              __typename
            }
            """
        assert_errors(
            query,
            [
                {"message": "Unknown type 'ID'.", "locations": [(2, 25)]},
                {"message": "Unknown type 'Float'.", "locations": [(2, 37)]},
                {"message": "Unknown type 'Int'.", "locations": [(2, 50)]},
            ],
            schema,
        )

    def describe_within_sdl():
        def use_standard_types():
            assert_sdl_valid(
                """
                type Query {
                  string: String
                  int: Int
                  float: Float
                  boolean: Boolean
                  id: ID
                  introspectionType: __EnumValue
                }
                """
            )

        def reference_types_defined_inside_the_same_document():
            assert_sdl_valid(
                """
                union SomeUnion = SomeObject | AnotherObject

                type SomeObject implements SomeInterface {
                someScalar(arg: SomeInputObject): SomeScalar
                }

                type AnotherObject {
                foo(arg: SomeInputObject): String
                }

                type SomeInterface {
                someScalar(arg: SomeInputObject): SomeScalar
                }

                input SomeInputObject {
                someScalar: SomeScalar
                }

                scalar SomeScalar

                type RootQuery {
                someInterface: SomeInterface
                someUnion: SomeUnion
                someScalar: SomeScalar
                someObject: SomeObject
                }

                schema {
                query: RootQuery
                }
                """
            )

        def unknown_type_references():
            assert_sdl_errors(
                """
                type A
                type B

                type SomeObject implements C {
                  e(d: D): E
                }

                union SomeUnion = F | G

                interface SomeInterface {
                  i(h: H): I
                }

                input SomeInput {
                  j: J
                }

                directive @SomeDirective(k: K) on QUERY

                schema {
                  query: L
                  mutation: M
                  subscription: N
                }
                """,
                [
                    {
                        "message": "Unknown type 'C'. Did you mean 'A' or 'B'?",
                        "locations": [(5, 44)],
                    },
                    {
                        "message": "Unknown type 'D'. Did you mean 'A', 'B', or 'ID'?",
                        "locations": [(6, 24)],
                    },
                    {
                        "message": "Unknown type 'E'. Did you mean 'A' or 'B'?",
                        "locations": [(6, 28)],
                    },
                    {
                        "message": "Unknown type 'F'. Did you mean 'A' or 'B'?",
                        "locations": [(9, 35)],
                    },
                    {
                        "message": "Unknown type 'G'. Did you mean 'A' or 'B'?",
                        "locations": [(9, 39)],
                    },
                    {
                        "message": "Unknown type 'H'. Did you mean 'A' or 'B'?",
                        "locations": [(12, 24)],
                    },
                    {
                        "message": "Unknown type 'I'. Did you mean 'A', 'B', or 'ID'?",
                        "locations": [(12, 28)],
                    },
                    {
                        "message": "Unknown type 'J'. Did you mean 'A' or 'B'?",
                        "locations": [(16, 22)],
                    },
                    {
                        "message": "Unknown type 'K'. Did you mean 'A' or 'B'?",
                        "locations": [(19, 45)],
                    },
                    {
                        "message": "Unknown type 'L'. Did you mean 'A' or 'B'?",
                        "locations": [(22, 26)],
                    },
                    {
                        "message": "Unknown type 'M'. Did you mean 'A' or 'B'?",
                        "locations": [(23, 29)],
                    },
                    {
                        "message": "Unknown type 'N'. Did you mean 'A' or 'B'?",
                        "locations": [(24, 33)],
                    },
                ],
            )

        def does_not_consider_non_type_definitions():
            assert_sdl_errors(
                """
                query Foo { __typename }
                fragment Foo on Query { __typename }
                directive @Foo on QUERY

                type Query {
                  foo: Foo
                }
                """,
                [{"message": "Unknown type 'Foo'.", "locations": [(7, 24)]}],
            )

        def reference_standard_types_inside_extension_document():
            schema = build_schema("type Foo")
            sdl = """
                type SomeType {
                  string: String
                  int: Int
                  float: Float
                  boolean: Boolean
                  id: ID
                  introspectionType: __EnumValue
                }
                """

            assert_sdl_valid(sdl, schema=schema)

        def reference_types_inside_extension_document():
            schema = build_schema("type Foo")
            sdl = """
                type QueryRoot {
                  foo: Foo
                  bar: Bar
                }

                scalar Bar

                schema {
                  query: QueryRoot
                }
                """

            assert_sdl_valid(sdl, schema=schema)

        def unknown_type_references_inside_extension_document():
            schema = build_schema("type A")
            sdl = """
                type B

                type SomeObject implements C {
                  e(d: D): E
                }

                union SomeUnion = F | G

                interface SomeInterface {
                  i(h: H): I
                }

                input SomeInput {
                  j: J
                }

                directive @SomeDirective(k: K) on QUERY

                schema {
                  query: L
                  mutation: M
                  subscription: N
                }
                """

            assert_sdl_errors(
                sdl,
                [
                    {
                        "message": "Unknown type 'C'. Did you mean 'A' or 'B'?",
                        "locations": [(4, 44)],
                    },
                    {
                        "message": "Unknown type 'D'. Did you mean 'A', 'B', or 'ID'?",
                        "locations": [(5, 24)],
                    },
                    {
                        "message": "Unknown type 'E'. Did you mean 'A' or 'B'?",
                        "locations": [(5, 28)],
                    },
                    {
                        "message": "Unknown type 'F'. Did you mean 'A' or 'B'?",
                        "locations": [(8, 35)],
                    },
                    {
                        "message": "Unknown type 'G'. Did you mean 'A' or 'B'?",
                        "locations": [(8, 39)],
                    },
                    {
                        "message": "Unknown type 'H'. Did you mean 'A' or 'B'?",
                        "locations": [(11, 24)],
                    },
                    {
                        "message": "Unknown type 'I'. Did you mean 'A', 'B', or 'ID'?",
                        "locations": [(11, 28)],
                    },
                    {
                        "message": "Unknown type 'J'. Did you mean 'A' or 'B'?",
                        "locations": [(15, 22)],
                    },
                    {
                        "message": "Unknown type 'K'. Did you mean 'A' or 'B'?",
                        "locations": [(18, 45)],
                    },
                    {
                        "message": "Unknown type 'L'. Did you mean 'A' or 'B'?",
                        "locations": [(21, 26)],
                    },
                    {
                        "message": "Unknown type 'M'. Did you mean 'A' or 'B'?",
                        "locations": [(22, 29)],
                    },
                    {
                        "message": "Unknown type 'N'. Did you mean 'A' or 'B'?",
                        "locations": [(23, 33)],
                    },
                ],
                schema,
            )



from functools import partial

from graphql.validation import ExecutableDefinitionsRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, ExecutableDefinitionsRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_executable_definitions():
    def with_only_operation():
        assert_valid(
            """
            query Foo {
              dog {
                name
              }
            }
            """
        )

    def with_operation_and_fragment():
        assert_valid(
            """
            query Foo {
              dog {
                name
                ...Frag
              }
            }

            fragment Frag on Dog {
              name
            }
            """
        )

    def with_type_definition():
        assert_errors(
            """
            query Foo {
              dog {
                name
              }
            }

            type Cow {
              name: String
            }

            extend type Dog {
              color: String
            }
            """,
            [
                {
                    "message": "The 'Cow' definition is not executable.",
                    "locations": [(8, 13)],
                },
                {
                    "message": "The 'Dog' definition is not executable.",
                    "locations": [(12, 13)],
                },
            ],
        )

    def with_schema_definition():
        assert_errors(
            """
            schema {
              query: Query
            }

            type Query {
              test: String
            }

            extend schema @directive
            """,
            [
                {
                    "message": "The schema definition is not executable.",
                    "locations": [(2, 13)],
                },
                {
                    "message": "The 'Query' definition is not executable.",
                    "locations": [(6, 13)],
                },
                {
                    "message": "The schema definition is not executable.",
                    "locations": [(10, 13)],
                },
            ],
        )



from asyncio import sleep
from inspect import isawaitable

from pytest import mark, raises

from graphql.pyutils import SimplePubSub


def describe_simple_pub_sub():
    @mark.asyncio
    async def subscribe_async_iterator_mock():
        pubsub = SimplePubSub()
        iterator = pubsub.get_subscriber()

        # Queue up publishes
        assert pubsub.emit("Apple") is True
        assert pubsub.emit("Banana") is True

        # Read payloads
        assert await iterator.__anext__() == "Apple"
        assert await iterator.__anext__() == "Banana"

        # Read ahead
        i3 = await iterator.__anext__()
        assert isawaitable(i3)
        i4 = await iterator.__anext__()
        assert isawaitable(i4)

        # Publish
        assert pubsub.emit("Coconut") is True
        assert pubsub.emit("Durian") is True

        # Await out of order to get correct results
        assert await i4 == "Durian"
        assert await i3 == "Coconut"

        # Read ahead
        i5 = iterator.__anext__()

        # Terminate queue
        await iterator.aclose()

        # Publish is not caught after terminate
        assert pubsub.emit("Fig") is False

        # Find that cancelled read-ahead got a "done" result
        with raises(StopAsyncIteration):
            await i5

        # And next returns empty completion value
        with raises(StopAsyncIteration):
            await iterator.__anext__()

    @mark.asyncio
    async def iterator_aclose_empties_push_queue():
        pubsub = SimplePubSub()
        assert not pubsub.subscribers
        iterator = pubsub.get_subscriber()
        assert len(pubsub.subscribers) == 1
        assert iterator.listening
        for value in range(3):
            pubsub.emit(value)
        await sleep(0)
        assert iterator.push_queue.qsize() == 3
        assert iterator.pull_queue.qsize() == 0
        await iterator.aclose()
        assert not pubsub.subscribers
        assert iterator.push_queue.qsize() == 0
        assert iterator.pull_queue.qsize() == 0
        assert not iterator.listening

    @mark.asyncio
    async def iterator_aclose_empties_pull_queue():
        pubsub = SimplePubSub()
        assert not pubsub.subscribers
        iterator = pubsub.get_subscriber()
        assert len(pubsub.subscribers) == 1
        assert iterator.listening
        for _n in range(3):
            await iterator.__anext__()
        assert iterator.push_queue.qsize() == 0
        assert iterator.pull_queue.qsize() == 3
        await iterator.aclose()
        assert not pubsub.subscribers
        assert iterator.push_queue.qsize() == 0
        assert iterator.pull_queue.qsize() == 0
        assert not iterator.listening

    @mark.asyncio
    async def iterator_aclose_is_idempotent():
        pubsub = SimplePubSub()
        iterator = pubsub.get_subscriber()
        assert iterator.listening
        for n in range(3):
            await iterator.aclose()
            assert not iterator.listening



from typing import (
    Any,
    ByteString,
    Collection,
    Iterable,
    Mapping,
    Text,
    ValuesView,
)

__all__ = ["is_collection", "is_iterable"]

collection_types: Any = Collection
if not isinstance({}.values(), Collection):  # Python < 3.7.2
    collection_types = (Collection, ValuesView)
iterable_types: Any = Iterable
not_iterable_types: Any = (ByteString, Mapping, Text)


def is_collection(value: Any) -> bool:
    """Check if value is a collection, but not a string or a mapping."""
    return isinstance(value, collection_types) and not isinstance(
        value, not_iterable_types
    )


def is_iterable(value: Any) -> bool:
    """Check if value is an iterable, but not a string or a mapping."""
    return isinstance(value, iterable_types) and not isinstance(
        value, not_iterable_types
    )



from copy import deepcopy

from pytest import raises

from graphql.language import (
    DirectiveLocation,
    SchemaDefinitionNode,
    SchemaExtensionNode,
    TypeDefinitionNode,
    TypeExtensionNode,
)
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLDirective,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLInputObjectType,
    GraphQLInputField,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLString,
    GraphQLType,
    specified_directives,
)
from graphql.utilities import build_schema, lexicographic_sort_schema, print_schema

from ..utils import dedent


def describe_type_system_schema():
    def define_sample_schema():
        BlogImage = GraphQLObjectType(
            "Image",
            {
                "url": GraphQLField(GraphQLString),
                "width": GraphQLField(GraphQLInt),
                "height": GraphQLField(GraphQLInt),
            },
        )

        BlogArticle: GraphQLObjectType

        BlogAuthor = GraphQLObjectType(
            "Author",
            lambda: {
                "id": GraphQLField(GraphQLString),
                "name": GraphQLField(GraphQLString),
                "pic": GraphQLField(
                    BlogImage,
                    args={
                        "width": GraphQLArgument(GraphQLInt),
                        "height": GraphQLArgument(GraphQLInt),
                    },
                ),
                "recentArticle": GraphQLField(BlogArticle),
            },
        )

        BlogArticle = GraphQLObjectType(
            "Article",
            lambda: {
                "id": GraphQLField(GraphQLString),
                "isPublished": GraphQLField(GraphQLBoolean),
                "author": GraphQLField(BlogAuthor),
                "title": GraphQLField(GraphQLString),
                "body": GraphQLField(GraphQLString),
            },
        )

        BlogQuery = GraphQLObjectType(
            "Query",
            {
                "article": GraphQLField(
                    BlogArticle, args={"id": GraphQLArgument(GraphQLString)}
                ),
                "feed": GraphQLField(GraphQLList(BlogArticle)),
            },
        )

        BlogMutation = GraphQLObjectType(
            "Mutation", {"writeArticle": GraphQLField(BlogArticle)}
        )

        BlogSubscription = GraphQLObjectType(
            "Subscription",
            {
                "articleSubscribe": GraphQLField(
                    args={"id": GraphQLArgument(GraphQLString)}, type_=BlogArticle
                )
            },
        )

        schema = GraphQLSchema(
            BlogQuery,
            BlogMutation,
            BlogSubscription,
            description="Sample schema",
        )

        kwargs = schema.to_kwargs()
        types = kwargs.pop("types")
        assert types == tuple(schema.type_map.values())
        assert kwargs == {
            "query": BlogQuery,
            "mutation": BlogMutation,
            "subscription": BlogSubscription,
            "directives": specified_directives,
            "description": "Sample schema",
            "extensions": {},
            "ast_node": None,
            "extension_ast_nodes": (),
            "assume_valid": False,
        }

        assert print_schema(schema) == dedent(
            '''
            """Sample schema"""
            schema {
              query: Query
              mutation: Mutation
              subscription: Subscription
            }

            type Query {
              article(id: String): Article
              feed: [Article]
            }

            type Article {
              id: String
              isPublished: Boolean
              author: Author
              title: String
              body: String
            }

            type Author {
              id: String
              name: String
              pic(width: Int, height: Int): Image
              recentArticle: Article
            }

            type Image {
              url: String
              width: Int
              height: Int
            }

            type Mutation {
              writeArticle: Article
            }

            type Subscription {
              articleSubscribe(id: String): Article
            }
            '''
        )

    def freezes_the_specified_directives():
        directives_list = [GraphQLDirective("SomeDirective", [])]
        schema = GraphQLSchema(directives=directives_list)
        assert isinstance(schema.directives, tuple)
        assert schema.directives == tuple(directives_list)
        directives_tuple = schema.directives
        schema = GraphQLSchema(directives=directives_tuple)
        assert schema.directives is directives_tuple

    def rejects_a_schema_with_incorrectly_typed_description():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLSchema(description=[])  # type: ignore
        assert str(exc_info.value) == "Schema description must be a string."

    def describe_type_map():
        def includes_interface_possible_types_in_the_type_map():
            SomeInterface = GraphQLInterfaceType("SomeInterface", {})
            SomeSubtype = GraphQLObjectType(
                "SomeSubtype", {}, interfaces=[SomeInterface]
            )
            schema = GraphQLSchema(
                query=GraphQLObjectType(
                    "Query", {"iface": GraphQLField(SomeInterface)}
                ),
                types=[SomeSubtype],
            )
            assert schema.type_map["SomeInterface"] is SomeInterface
            assert schema.type_map["SomeSubtype"] is SomeSubtype

            assert schema.is_sub_type(SomeInterface, SomeSubtype) is True

        def includes_interfaces_thunk_subtypes_in_the_type_map():
            AnotherInterface = GraphQLInterfaceType("AnotherInterface", {})
            SomeInterface = GraphQLInterfaceType(
                "SomeInterface", {}, interfaces=lambda: [AnotherInterface]
            )
            SomeSubtype = GraphQLObjectType(
                "SomeSubtype", {}, interfaces=lambda: [SomeInterface]
            )
            schema = GraphQLSchema(
                types=[SomeSubtype],
            )
            assert schema.type_map["SomeInterface"] is SomeInterface
            assert schema.type_map["AnotherInterface"] is AnotherInterface
            assert schema.type_map["SomeSubtype"] is SomeSubtype

        def includes_nested_input_objects_in_the_map():
            NestedInputObject = GraphQLInputObjectType("NestedInputObject", {})
            SomeInputObject = GraphQLInputObjectType(
                "SomeInputObject", {"nested": GraphQLInputField(NestedInputObject)}
            )

            schema = GraphQLSchema(
                GraphQLObjectType(
                    "Query",
                    {
                        "something": GraphQLField(
                            GraphQLString, {"input": GraphQLArgument(SomeInputObject)}
                        )
                    },
                )
            )
            assert schema.type_map["SomeInputObject"] is SomeInputObject
            assert schema.type_map["NestedInputObject"] is NestedInputObject

        def includes_input_types_only_used_in_directives():
            directive = GraphQLDirective(
                name="dir",
                locations=[DirectiveLocation.OBJECT],
                args={
                    "arg": GraphQLArgument(
                        GraphQLInputObjectType(
                            "Foo", {"field": GraphQLInputField(GraphQLString)}
                        )
                    ),
                    "argList": GraphQLArgument(
                        GraphQLList(
                            GraphQLInputObjectType(
                                "Bar", {"field": GraphQLInputField(GraphQLString)}
                            )
                        )
                    ),
                },
            )
            schema = GraphQLSchema(directives=[directive])
            assert "Foo" in schema.type_map
            assert "Bar" in schema.type_map

    def preserves_the_order_of_user_provided_types():
        a_type = GraphQLObjectType(
            "A", {"sub": GraphQLField(GraphQLScalarType("ASub"))}
        )
        z_type = GraphQLObjectType(
            "Z", {"sub": GraphQLField(GraphQLScalarType("ZSub"))}
        )
        query_type = GraphQLObjectType(
            "Query",
            {
                "a": GraphQLField(a_type),
                "z": GraphQLField(z_type),
                "sub": GraphQLField(GraphQLScalarType("QuerySub")),
            },
        )
        schema = GraphQLSchema(query_type, types=[z_type, query_type, a_type])

        type_names = list(schema.type_map)
        assert type_names == [
            "Z",
            "ZSub",
            "Query",
            "QuerySub",
            "A",
            "ASub",
            "Boolean",
            "String",
            "__Schema",
            "__Type",
            "__TypeKind",
            "__Field",
            "__InputValue",
            "__EnumValue",
            "__Directive",
            "__DirectiveLocation",
        ]

        # Also check that this order is stable
        copy_schema = GraphQLSchema(**schema.to_kwargs())
        assert list(copy_schema.type_map) == type_names

    def describe_validity():
        def describe_when_not_assumed_valid():
            def configures_the_schema_to_still_needing_validation():
                # noinspection PyProtectedMember
                assert GraphQLSchema(assume_valid=False).validation_errors is None

            def checks_the_configuration_for_mistakes():
                def query():
                    pass

                with raises(Exception):
                    # noinspection PyTypeChecker
                    GraphQLSchema(query)  # type: ignore
                with raises(Exception):
                    GraphQLSchema(types={})
                with raises(Exception):
                    GraphQLSchema(directives={})

            def check_that_query_mutation_and_subscription_are_graphql_types():
                directive = GraphQLDirective("foo", [])
                with raises(TypeError) as exc_info:
                    # noinspection PyTypeChecker
                    GraphQLSchema(query=directive)  # type: ignore
                assert str(exc_info.value) == "Expected query to be a GraphQL type."
                with raises(TypeError) as exc_info:
                    # noinspection PyTypeChecker
                    GraphQLSchema(mutation=directive)  # type: ignore
                assert str(exc_info.value) == (
                    "Expected mutation to be a GraphQL type."
                )
                with raises(TypeError) as exc_info:
                    # noinspection PyTypeChecker
                    GraphQLSchema(subscription=directive)  # type: ignore
                assert str(exc_info.value) == (
                    "Expected subscription to be a GraphQL type."
                )

    def describe_a_schema_must_contain_uniquely_named_types():
        def rejects_a_schema_which_redefines_a_built_in_type():
            FakeString = GraphQLScalarType("String")

            QueryType = GraphQLObjectType(
                "Query",
                {
                    "normal": GraphQLField(GraphQLString),
                    "fake": GraphQLField(FakeString),
                },
            )

            with raises(TypeError) as exc_info:
                GraphQLSchema(QueryType)
            msg = str(exc_info.value)
            assert msg == (
                "Schema must contain uniquely named types"
                " but contains multiple types named 'String'."
            )

        def rejects_a_schema_when_a_provided_type_has_no_name():
            query = GraphQLObjectType("Query", {"foo": GraphQLField(GraphQLString)})
            types = [GraphQLType(), query, GraphQLType()]

            with raises(TypeError) as exc_info:
                GraphQLSchema(query, types=types)  # type: ignore
            msg = str(exc_info.value)
            assert msg == (
                "One of the provided types for building the Schema is missing a name."
            )

        def rejects_a_schema_which_defines_an_object_twice():
            types = [
                GraphQLObjectType("SameName", {}),
                GraphQLObjectType("SameName", {}),
            ]

            with raises(TypeError) as exc_info:
                GraphQLSchema(types=types)
            msg = str(exc_info.value)
            assert msg == (
                "Schema must contain uniquely named types"
                " but contains multiple types named 'SameName'."
            )

        def rejects_a_schema_which_defines_fields_with_conflicting_types():
            fields: GraphQLFieldMap = {}
            QueryType = GraphQLObjectType(
                "Query",
                {
                    "a": GraphQLField(GraphQLObjectType("SameName", fields)),
                    "b": GraphQLField(GraphQLObjectType("SameName", fields)),
                },
            )

            with raises(TypeError) as exc_info:
                GraphQLSchema(QueryType)
            msg = str(exc_info.value)
            assert msg == (
                "Schema must contain uniquely named types"
                " but contains multiple types named 'SameName'."
            )

        def describe_when_assumed_valid():
            def configures_the_schema_to_have_no_errors():
                # noinspection PyProtectedMember
                assert GraphQLSchema(assume_valid=True).validation_errors == []

    def describe_ast_nodes():
        def accepts_a_scalar_type_with_ast_node_and_extension_ast_nodes():
            ast_node = SchemaDefinitionNode()
            extension_ast_nodes = [SchemaExtensionNode()]
            schema = GraphQLSchema(
                GraphQLObjectType("Query", {}),
                ast_node=ast_node,
                extension_ast_nodes=extension_ast_nodes,
            )
            assert schema.ast_node is ast_node
            assert schema.extension_ast_nodes == tuple(extension_ast_nodes)

        def rejects_a_schema_with_an_incorrect_ast_node():
            with raises(TypeError) as exc_info:
                # noinspection PyTypeChecker
                GraphQLSchema(
                    GraphQLObjectType("Query", {}),
                    ast_node=TypeDefinitionNode(),  # type: ignore
                )
            msg = str(exc_info.value)
            assert msg == "Schema AST node must be a SchemaDefinitionNode."

        def rejects_a_scalar_type_with_incorrect_extension_ast_nodes():
            with raises(TypeError) as exc_info:
                # noinspection PyTypeChecker
                GraphQLSchema(
                    GraphQLObjectType("Query", {}),
                    extension_ast_nodes=[TypeExtensionNode()],  # type: ignore
                )
            assert str(exc_info.value) == (
                "Schema extension AST nodes must be specified"
                " as a collection of SchemaExtensionNode instances."
            )

    def can_deep_copy_a_schema():
        source = """
            schema {
              query: Farm
              mutation: Work
            }

            type Cow {
              id: ID!
              name: String
              moos: Boolean
            }

            type Pig {
              id: ID!
              name: String
              oink: Boolean
            }

            union Animal = Cow | Pig

            enum Food {
              CORN
              FRUIT
            }

            input Feed {
              amount: Float
              type: Food
            }

            type Farm {
              animals: [Animal!]!
            }

            type Work {
              feed(feed: Feed): Boolean
            }
        """
        schema = build_schema(source)
        schema_copy = deepcopy(schema)

        for name in ("Cow", "Pig", "Animal", "Food", "Feed", "Farm", "Work"):
            assert schema.get_type(name) is not schema_copy.get_type(name)

        assert print_schema(lexicographic_sort_schema(schema)) == print_schema(
            lexicographic_sort_schema(schema_copy)
        )



from graphql.language import parse_value, print_ast
from graphql.utilities.sort_value_node import sort_value_node


def describe_sort_value_node():
    def _expect_sorted_value(source: str, expected: str) -> None:
        assert print_ast(sort_value_node(parse_value(source))) == expected

    def do_not_change_non_object_values():
        _expect_sorted_value("1", "1")
        _expect_sorted_value("3.14", "3.14")
        _expect_sorted_value("null", "null")
        _expect_sorted_value("true", "true")
        _expect_sorted_value("false", "false")
        _expect_sorted_value('"cba"', '"cba"')
        _expect_sorted_value(
            '[1, 3.14, null, false, "cba"]', '[1, 3.14, null, false, "cba"]'
        )
        _expect_sorted_value(
            '[[1, 3.14, null, false, "cba"]]', '[[1, 3.14, null, false, "cba"]]'
        )

    def sort_input_object_fields():
        _expect_sorted_value("{ b: 2, a: 1 }", "{a: 1, b: 2}")
        _expect_sorted_value("{ a: { c: 3, b: 2 } }", "{a: {b: 2, c: 3}}")
        _expect_sorted_value(
            "[{ b: 2, a: 1 }, { d: 4, c: 3}]",
            "[{a: 1, b: 2}, {c: 3, d: 4}]",
        )
        _expect_sorted_value(
            "{ b: { g: 7, f: 6 }, c: 3 , a: { d: 4, e: 5 } }",
            "{a: {d: 4, e: 5}, b: {f: 6, g: 7}, c: 3}",
        )
